"""Module for Local TGI Serving"""

from __future__ import absolute_import

import requests
import logging
from pathlib import Path
from docker.types import DeviceRequest
from sagemaker.core.helper.session_helper import Session
from sagemaker.core import fw_utils
from sagemaker.core.s3.utils import determine_bucket_and_prefix, parse_s3_url, s3_path_join
from sagemaker.core.s3 import S3Uploader
from sagemaker.core.local.local_session import get_docker_host
from sagemaker.core.common_utils import _is_s3_uri

MODE_DIR_BINDING = "/opt/ml/model/"
_SHM_SIZE = "2G"
_DEFAULT_ENV_VARS = {
    "HF_HOME": "/opt/ml/model/",
    "HUGGINGFACE_HUB_CACHE": "/opt/ml/model/",
}

logger = logging.getLogger(__name__)


class LocalTgiServing:
    """Placeholder docstring"""

    def _start_tgi_serving(
        self,
        client: object,
        image: str,
        model_path: str,
        secret_key: str,
        env_vars: dict,
        jumpstart: bool,
    ):
        """Placeholder docstring"""
        if jumpstart:
            self.container = client.containers.run(
                image,
                ["--model-id", MODE_DIR_BINDING],
                shm_size=_SHM_SIZE,
                device_requests=[DeviceRequest(count=-1, capabilities=[["gpu"]])],
                network_mode="host",
                detach=True,
                auto_remove=True,
                volumes={
                    Path(model_path).joinpath("code"): {
                        "bind": MODE_DIR_BINDING,
                        "mode": "rw",
                    },
                },
                environment=env_vars,
            )
        else:
            self.container = client.containers.run(
                image,
                shm_size=_SHM_SIZE,
                device_requests=[DeviceRequest(count=-1, capabilities=[["gpu"]])],
                network_mode="host",
                detach=True,
                auto_remove=True,
                volumes={
                    Path(model_path).joinpath("code"): {
                        "bind": MODE_DIR_BINDING,
                        "mode": "rw",
                    },
                },
                environment=_update_env_vars(env_vars),
            )

    def _invoke_tgi_serving(self, request: object, content_type: str, accept: str):
        """Placeholder docstring"""
        try:
            response = requests.post(
                f"http://{get_docker_host()}:8080/generate",
                data=request,
                headers={"Content-Type": content_type, "Accept": accept},
                timeout=600,
            )
            response.raise_for_status()
            return response.content
        except Exception as e:
            raise Exception("Unable to send request to the local container server") from e


class SageMakerTgiServing:
    """Placeholder docstring"""

    def _upload_tgi_artifacts(
        self,
        model_path: str,
        sagemaker_session: Session,
        jumpstart: bool,
        s3_model_data_url: str = None,
        image: str = None,
        env_vars: dict = None,
        should_upload_artifacts: bool = False,
    ):
        model_data_url = None
        if _is_s3_uri(model_path):
            model_data_url = model_path
        elif should_upload_artifacts:
            if s3_model_data_url:
                bucket, key_prefix = parse_s3_url(url=s3_model_data_url)
            else:
                bucket, key_prefix = None, None

            code_key_prefix = fw_utils.model_code_key_prefix(key_prefix, None, image)

            bucket, code_key_prefix = determine_bucket_and_prefix(
                bucket=bucket, key_prefix=code_key_prefix, sagemaker_session=sagemaker_session
            )

            code_dir = Path(model_path).joinpath("code")

            s3_location = s3_path_join("s3://", bucket, code_key_prefix, "code")

            logger.debug("Uploading TGI Model Resources uncompressed to: %s", s3_location)

            model_data_url = S3Uploader.upload(
                str(code_dir),
                s3_location,
                None,
                sagemaker_session,
            )

        model_data = (
            {
                "S3DataSource": {
                    "CompressionType": "None",
                    "S3DataType": "S3Prefix",
                    "S3Uri": model_data_url + "/",
                }
            }
            if model_data_url
            else None
        )
        if jumpstart:
            return (model_data, {})
        return (model_data, _update_env_vars(env_vars))


def _update_env_vars(env_vars: dict) -> dict:
    """Placeholder docstring"""
    updated_env_vars = {}
    updated_env_vars.update(_DEFAULT_ENV_VARS)
    if env_vars:
        updated_env_vars.update(env_vars)
    return updated_env_vars



"""Placeholder docerting"""

from __future__ import absolute_import
import uuid
import logging
import importlib
import platform

from sagemaker.core import fw_utils
from sagemaker.core.helper.session_helper import Session
from sagemaker.core.common_utils import _is_s3_uri
from sagemaker.serve.utils.uploader import upload
from sagemaker.core.s3.utils import determine_bucket_and_prefix, parse_s3_url
from sagemaker.core.local.local_session import get_docker_host
import docker
from docker.types import DeviceRequest

logger = logging.getLogger(__name__)

# TODO: automatically update memory size
_SHM_SIZE = "2G"


class LocalTritonServer:
    """Placeholder docstring"""

    def __init__(self) -> None:
        self.triton_client = None

    def _start_triton_server(
        self,
        docker_client: docker.DockerClient,
        model_path: str,
        secret_key: str,
        image_uri: str,
        env_vars: dict,
    ):
        """Placeholder docstring"""
        self.container_name = "triton" + uuid.uuid1().hex
        model_repository = model_path + "/model_repository"
        env_vars.update(
            {
                "TRITON_MODEL_DIR": "/models/model",
                "SAGEMAKER_SERVE_SECRET_KEY": secret_key,
                "LOCAL_PYTHON": platform.python_version(),
            }
        )

        if "cpu" not in image_uri:
            self.container = docker_client.containers.run(
                image=image_uri,
                command=["tritonserver", "--model-repository=/models"],
                shm_size=_SHM_SIZE,
                device_requests=[DeviceRequest(count=-1, capabilities=[["gpu"]])],
                network_mode="host",
                detach=True,
                auto_remove=True,
                volumes={model_repository: {"bind": "/models", "mode": "rw"}},
                environment=env_vars,
            )

        else:
            self.container = docker_client.containers.run(
                image=image_uri,
                command=["tritonserver", "--model-repository=/models"],
                shm_size=_SHM_SIZE,
                network_mode="host",
                detach=True,
                auto_remove=True,
                volumes={model_repository: {"bind": "/models", "mode": "rw"}},
                environment=env_vars,
            )

    def _invoke_triton_server(self, payload, *args, **kwargs):
        """Placeholder docstring"""
        httpClient = importlib.import_module("tritonclient.http")

        if not self.triton_client:
            self.triton_client = httpClient.InferenceServerClient(url=f"{get_docker_host()}:8000")

        payload = self.schema_builder.input_serializer.serialize(payload)
        dtype = self.schema_builder._input_triton_dtype.split("_")[-1]

        input_request = httpClient.InferInput("input_1", payload.shape, datatype=dtype)
        input_request.set_data_from_numpy(payload, binary_data=True)

        response = self.triton_client.infer(model_name="model", inputs=[input_request])
        response_name = response.get_response().get("outputs")[0].get("name")

        return self.schema_builder.output_deserializer.deserialize(response.as_numpy(response_name))


class SageMakerTritonServer:
    """Placeholder docstring"""

    def __init__(self) -> None:
        pass

    def _upload_triton_artifacts(
        self,
        model_path: str,
        sagemaker_session: Session,
        secret_key: str,
        s3_model_data_url: str = None,
        image: str = None,
        should_upload_artifacts: bool = False,
    ):
        """Tar triton artifacts and upload to s3"""
        s3_upload_path = None
        if _is_s3_uri(model_path):
            s3_upload_path = model_path
        elif should_upload_artifacts:
            if s3_model_data_url:
                bucket, key_prefix = parse_s3_url(url=s3_model_data_url)
            else:
                bucket, key_prefix = None, None

            code_key_prefix = fw_utils.model_code_key_prefix(key_prefix, None, image)

            bucket, code_key_prefix = determine_bucket_and_prefix(
                bucket=bucket, key_prefix=code_key_prefix, sagemaker_session=sagemaker_session
            )

            logger.debug(
                "Uploading the model resources to bucket=%s, key_prefix=%s.",
                bucket,
                code_key_prefix,
            )
            model_repository = model_path + "/model_repository"
            s3_upload_path = upload(sagemaker_session, model_repository, bucket, code_key_prefix)
            logger.debug("Model resources uploaded to: %s", s3_upload_path)

        env_vars = {
            "SAGEMAKER_TRITON_DEFAULT_MODEL_NAME": "model",
            "TRITON_MODEL_DIR": "/opt/ml/model/model",
            "SAGEMAKER_SERVE_SECRET_KEY": secret_key,
            "LOCAL_PYTHON": platform.python_version(),
        }
        return s3_upload_path, env_vars



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""A class for WaiterConfig used in async inference

Use it when using async inference and wait for the result.
"""

from __future__ import absolute_import


class WaiterConfig(object):
    """Configuration object passed in when using async inference and wait for the result."""

    def __init__(
        self,
        max_attempts=60,
        delay=15,
    ):
        """Initialize a WaiterConfig object that provides parameters to control waiting behavior.

        Args:
            max_attempts (int): The maximum number of attempts to be made. If the max attempts is
            exceeded, Amazon SageMaker will raise ``PollingTimeoutError``. (Default: 60)
            delay (int): The amount of time in seconds to wait between attempts. (Default: 15)
        """

        self.max_attempts = max_attempts
        self.delay = delay

    def _to_request_dict(self):
        """Generates a dictionary using the parameters provided to the class."""
        waiter_dict = {
            "Delay": self.delay,
            "MaxAttempts": self.max_attempts,
        }

        return waiter_dict



"""Validates the integrity of pickled file with HMAC signing."""

from __future__ import absolute_import
import secrets
import hmac
import hashlib
import os
from pathlib import Path

from sagemaker.core.remote_function.core.serialization import _MetaData


def generate_secret_key(nbytes: int = 32) -> str:
    """Generates secret key"""
    return secrets.token_hex(nbytes)


def compute_hash(buffer: bytes, secret_key: str) -> str:
    """Compute hash value using HMAC"""
    return hmac.new(secret_key.encode(), msg=buffer, digestmod=hashlib.sha256).hexdigest()


def perform_integrity_check(buffer: bytes, metadata_path: Path):
    """Validates the integrity of bytes by comparing the hash value"""
    secret_key = os.environ.get("SAGEMAKER_SERVE_SECRET_KEY")
    actual_hash_value = compute_hash(buffer=buffer, secret_key=secret_key)

    if not Path.exists(metadata_path):
        raise ValueError("Path to metadata.json does not exist")

    with open(str(metadata_path), "rb") as md:
        expected_hash_value = _MetaData.from_json(md.read()).sha256_hash

    if not hmac.compare_digest(expected_hash_value, actual_hash_value):
        raise ValueError("Integrity check for the serialized function or data failed.")



"""Module for In_process Serving"""

from __future__ import absolute_import

import requests
import logging

logger = logging.getLogger(__name__)


class InProcessServing:
    """In Process Mode server instance"""

    def _start_serving(self):
        """Initializes the start of the server"""
        from sagemaker.serve.model_server.in_process_model_server.app import InProcessServer

        self.server = InProcessServer(
            inference_spec=self.inference_spec, model=self.model, schema_builder=self.schema_builder
        )
        self.server.start_server()

    def _stop_serving(self):
        """Stops the server"""
        self.server.stop_server()

    def _invoke_serving(self, request: object, content_type: str, accept: str):
        """Placeholder docstring"""
        try:
            response = requests.post(
                f"http://{self.server.host}:{self.server.port}/invoke",
                data=request,
                headers={"Content-Type": content_type, "Accept": accept},
                timeout=600,
            )
            response.raise_for_status()

            return response.content
        except Exception as e:
            if "Connection refused" in str(e):
                raise Exception(
                    "Unable to send request to the local server: Connection refused."
                ) from e
            raise Exception("Unable to send request to the local container server %s", str(e))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""A class for AsyncInferenceResponse"""

from __future__ import print_function, absolute_import

from botocore.exceptions import ClientError
from sagemaker.core.s3 import parse_s3_url
from sagemaker.serve.async_inference import WaiterConfig
from sagemaker.core.exceptions import (
    ObjectNotExistedError,
    UnexpectedClientError,
    AsyncInferenceModelError,
)


class AsyncInferenceResponse(object):
    """Response from Async Inference endpoint

    This response object provides a method to check for an async inference result in the
    Amazon S3 output path specified. If result object exists in that path, get and return
    the result
    """

    def __init__(
        self,
        predictor_async,
        output_path,
        failure_path,
    ):
        """Initialize an AsyncInferenceResponse object.

        AsyncInferenceResponse can help users to get async inference result
        from the Amazon S3 output path

        Args:
            predictor_async (sagemaker.predictor.AsyncPredictor): The ``AsyncPredictor``
                that return this response.
            output_path (str): The Amazon S3 location that endpoints upload inference responses
                to.
            failure_path (str): The Amazon S3 location that endpoints upload model errors
                for failed requests.
        """
        self.predictor_async = predictor_async
        self.output_path = output_path
        self._result = None
        self.failure_path = failure_path

    def get_result(
        self,
        waiter_config=None,
    ):
        """Get async inference result in the Amazon S3 output path specified

        Args:
            waiter_config (sagemaker.async_inference.waiter_config.WaiterConfig): Configuration
                for the waiter. The pre-defined value for the delay between poll is 15 seconds
                and the default max attempts is 60
        Raises:
            ValueError: If a wrong type of object is provided as ``waiter_config``
        Returns:
            object: Inference result in the given Amazon S3 output path. If a deserializer was
                specified when creating the AsyncPredictor, the result of the deserializer is
                returned. Otherwise the response returns the sequence of bytes
                as is.
        """
        if waiter_config is not None and not isinstance(waiter_config, WaiterConfig):
            raise ValueError("waiter_config should be a WaiterConfig object")

        if self._result is None:
            if waiter_config is None:
                self._result = self._get_result_from_s3(self.output_path, self.failure_path)
            else:
                self._result = self.predictor_async._wait_for_output(
                    self.output_path, self.failure_path, waiter_config
                )
        return self._result

    def _get_result_from_s3(self, output_path, failure_path):
        """Retrieve output based on the presense of failure_path"""
        if failure_path is not None:
            return self._get_result_from_s3_output_failure_paths(output_path, failure_path)

        return self._get_result_from_s3_output_path(output_path)

    def _get_result_from_s3_output_path(self, output_path):
        """Get inference result from the output Amazon S3 path"""
        bucket, key = parse_s3_url(output_path)
        try:
            response = self.predictor_async.s3_client.get_object(Bucket=bucket, Key=key)
            return self.predictor_async.predictor._handle_response(response)
        except ClientError as ex:
            if ex.response["Error"]["Code"] == "NoSuchKey":
                raise ObjectNotExistedError(
                    message="Inference could still be running",
                    output_path=output_path,
                )
            raise UnexpectedClientError(
                message=ex.response["Error"]["Message"],
            )

    def _get_result_from_s3_output_failure_paths(self, output_path, failure_path):
        """Get inference result from the output & failure Amazon S3 path"""
        bucket, key = parse_s3_url(output_path)
        try:
            response = self.predictor_async.s3_client.get_object(Bucket=bucket, Key=key)
            return self.predictor_async.predictor._handle_response(response)
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchKey":
                try:
                    failure_bucket, failure_key = parse_s3_url(failure_path)
                    failure_response = self.predictor_async.s3_client.get_object(
                        Bucket=failure_bucket, Key=failure_key
                    )
                    failure_response = self.predictor_async.predictor._handle_response(
                        failure_response
                    )
                    raise AsyncInferenceModelError(message=failure_response)
                except ClientError as ex:
                    if ex.response["Error"]["Code"] == "NoSuchKey":
                        raise ObjectNotExistedError(
                            message="Inference could still be running", output_path=output_path
                        )
                    raise UnexpectedClientError(message=ex.response["Error"]["Message"])
            raise UnexpectedClientError(message=e.response["Error"]["Message"])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Constants and enums for SageMaker ModelBuilder and serving functionality.

This module defines:
- Framework enum for ML framework identification
- Supported model servers and local modes
- Default serializers and deserializers by framework
- Configuration constants for model serving

Example:
    Using Framework enum::
    
        from sagemaker.serve.constants import Framework, DEFAULT_SERIALIZERS_BY_FRAMEWORK
        
        # Get serializers for PyTorch
        serializer, deserializer = DEFAULT_SERIALIZERS_BY_FRAMEWORK[Framework.PYTORCH]
"""
from __future__ import absolute_import, annotations

# Standard library imports
from enum import Enum
from typing import Dict, Set, Tuple

# SageMaker imports
from sagemaker.serve.mode.function_pointers import Mode
from sagemaker.serve.utils.types import ModelServer
from sagemaker.core.deserializers import (
    CSVDeserializer,
    JSONDeserializer,
    NumpyDeserializer,
    RecordDeserializer,
)
from sagemaker.core.serializers import (
    JSONSerializer,
    LibSVMSerializer,
    NumpySerializer,
    RecordSerializer,
    TorchTensorSerializer,
)


# ========================================
# Mode and Server Constants
# ========================================

LOCAL_MODES: list[Mode] = [Mode.LOCAL_CONTAINER, Mode.IN_PROCESS]

SUPPORTED_MODEL_SERVERS: Set[ModelServer] = {
    ModelServer.TORCHSERVE,
    ModelServer.TRITON,
    ModelServer.DJL_SERVING,
    ModelServer.TENSORFLOW_SERVING,
    ModelServer.MMS,
    ModelServer.TGI,
    ModelServer.TEI,
    ModelServer.SMD,
}

# ========================================
# Framework Enum
# ========================================

class Framework(Enum):
    """Enumeration of supported ML frameworks for ModelBuilder.
    
    This enum provides standardized framework identifiers used throughout
    the ModelBuilder ecosystem for:
    - Framework detection from container images
    - Serializer/deserializer selection
    - Model server compatibility
    
    Example:
        Using framework enum::
        
            if detected_framework == Framework.PYTORCH:
                serializer, deserializer = DEFAULT_SERIALIZERS_BY_FRAMEWORK[Framework.PYTORCH]
    """
    XGBOOST = "XGBoost"
    LDA = "LDA"
    PYTORCH = "PyTorch"
    TENSORFLOW = "TensorFlow"
    MXNET = "MXNet"
    CHAINER = "Chainer"
    SKLEARN = "SKLearn"
    HUGGINGFACE = "HuggingFace"
    DJL = "DJL"
    SPARKML = "SparkML"
    NTM = "NTM"
    SMD = "SMD"


# ========================================
# Framework Serialization Mapping
# ========================================

DEFAULT_SERIALIZERS_BY_FRAMEWORK: Dict[Framework, Tuple] = {
    Framework.XGBOOST: (LibSVMSerializer(), CSVDeserializer()),
    Framework.LDA: (RecordSerializer(), RecordDeserializer()),
    Framework.PYTORCH: (TorchTensorSerializer(), JSONDeserializer()),
    Framework.TENSORFLOW: (NumpySerializer(), JSONDeserializer()),
    Framework.MXNET: (RecordSerializer(), JSONDeserializer()),
    Framework.CHAINER: (NumpySerializer(), JSONDeserializer()),
    Framework.SKLEARN: (NumpySerializer(), NumpyDeserializer()),
    Framework.HUGGINGFACE: (JSONSerializer(), JSONDeserializer()),
    Framework.DJL: (JSONSerializer(), JSONDeserializer()),
    Framework.SPARKML: (NumpySerializer(), JSONDeserializer()),
    Framework.NTM: (RecordSerializer(), JSONDeserializer()),
    Framework.SMD: (JSONSerializer(), JSONDeserializer()),
}




# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Utility functions for fetching model information from HuggingFace Hub"""
from __future__ import absolute_import
import json
import urllib.request
from json import JSONDecodeError
from urllib.error import HTTPError, URLError
import logging

logger = logging.getLogger(__name__)


def _get_model_config_properties_from_hf(model_id: str, hf_hub_token: str = None):
    """Placeholder docstring"""

    config_url = f"https://huggingface.co/{model_id}/raw/main/config.json"
    model_config = None
    try:
        if hf_hub_token:
            config_url = urllib.request.Request(
                config_url, headers={"Authorization": "Bearer " + hf_hub_token}
            )
        with urllib.request.urlopen(config_url) as response:
            model_config = json.load(response)
    except (HTTPError, URLError, TimeoutError, JSONDecodeError) as e:
        if "HTTP Error 401: Unauthorized" in str(e):
            raise ValueError(
                "Trying to access a gated/private HuggingFace model without valid credentials. "
                "Please provide a HUGGING_FACE_HUB_TOKEN in env_vars"
            )
        logger.warning(
            "Exception encountered while trying to read config file %s. " "Details: %s",
            config_url,
            e,
        )
    if not model_config:
        raise ValueError(
            f"Did not find a config.json or model_index.json file in huggingface hub for "
            f"{model_id}. Please make sure a config.json exists (or model_index.json for Stable "
            f"Diffusion Models) for this model in the huggingface hub"
        )
    return model_config



"""Save the object using cloudpickle"""

from __future__ import absolute_import
from typing import Any
from pathlib import Path
import cloudpickle

PKL_FILE_NAME = "serve.pkl"


def save_pkl(save_path: Path, obj: Any):
    """Save obj with cloudpickle under save_path"""
    if not save_path.exists():
        save_path.mkdir(parents=True)
    with open(save_path.joinpath(PKL_FILE_NAME), mode="wb") as file:
        cloudpickle.dump(obj, file)


def save_xgboost(save_path: Path, xgb_model: Any):
    """Save xgboost model to json format using save_model"""
    if not save_path.exists():
        save_path.mkdir(parents=True)
    xgb_model.save_model(str(save_path.joinpath("model.json")))

def save_sklearn(model_path: str, model: object) -> None:
    """Save sklearn model using joblib serialization."""
    import joblib
    import os
    from pathlib import Path
    
    # Ensure directory exists
    Path(model_path).mkdir(parents=True, exist_ok=True)
    
    model_file = os.path.join(model_path, "model.joblib")
    joblib.dump(model, model_file)


def load_xgboost_from_json(model_save_path: str, class_name: str):
    """Load xgboost model from json format"""
    try:
        kls = _get_class_from_name(class_name=class_name)
        xgb_model = kls()
        xgb_model.load_model(model_save_path)
        return xgb_model
    except Exception as e:
        raise ValueError(
            (
                "Unable to instantiate %s due to %s, please provide"
                "your custom code for loading the model with InferenceSpec"
            )
            % (class_name, e)
        )


def _get_class_from_name(class_name: str):
    """Given a full class name like xgboost.sklearn.XGBClassifier, return the class"""
    parts = class_name.split(".")
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)

    return m



"""Module for pulling logs from container"""

from __future__ import absolute_import
import logging
from threading import Thread
import queue
from datetime import datetime

from sagemaker.serve.utils.exceptions import (
    LocalModelLoadException,
    LocalModelOutOfMemoryException,
    LocalModelInvocationException,
    SkipTuningComboException,
)

logger = logging.getLogger(__name__)


def _get_logs(generator, logs, until):
    """Placeholder docstring"""
    now = datetime.now()
    try:
        for next_entry in generator:
            logs.put(next_entry, block=True, timeout=(until - now).total_seconds())

            now = datetime.now()
            if until < now:
                break
        logger.debug("Container logging done. All container logs processed.")
    except (StopIteration, ValueError):
        logger.debug("Container logging stopped. All container logs processed.")


def pull_logs(generator, stop, until, final_pull):
    """Placeholder docstring"""
    now = datetime.now()
    if until < now:
        return

    logs = queue.Queue(maxsize=10)
    log_puller = Thread(
        target=_get_logs,
        args=(
            generator,
            logs,
            until,
        ),
    )
    log_puller.start()

    while True:
        try:
            top = logs.get(block=True, timeout=(until - now).total_seconds())
            logger.debug(top)

            if "[INFO ]" in top and "OutOfMemoryError" in top:
                raise LocalModelOutOfMemoryException(top)
            if "CUDA out of memory. Tried to allocate" in top:
                raise LocalModelOutOfMemoryException(top)
            if "ai.djl.engine.EngineException: OOM" in top:
                raise LocalModelOutOfMemoryException(top)
            if "4xx.Count:1" in top or "5xx.Count:1" in top:
                raise LocalModelInvocationException(top)
            if "[ERROR]" in top or "Failed register workflow" in top:
                raise LocalModelLoadException(top)
            if "Address already in use" in top:
                raise LocalModelLoadException(top)
            if "not compatible with sharding" in top:
                raise SkipTuningComboException(top)
        except (
            LocalModelLoadException,
            LocalModelOutOfMemoryException,
            LocalModelInvocationException,
        ) as e:
            stop()
            log_puller.join(timeout=5)
            raise e
        except queue.Empty:
            now = datetime.now()
            if until < now:
                if not final_pull:
                    return

                stop()
                # allow logging thread some time to cleanup
                log_puller.join(timeout=5)
                if log_puller.is_alive():
                    raise Exception("Logging thread not terminating")
                return



"""Load a pickled object to detect the dependencies it requires"""

from __future__ import absolute_import
from pathlib import Path
from typing import List
import email.parser
import email.policy
import json
import inspect
import itertools
import subprocess
import sys
import tqdm

# non native imports. Ideally add as little as possible here
# because it will add to requirements.txt
import cloudpickle
import boto3

pipcmd = [sys.executable, "-m", "pip", "--disable-pip-version-check"]


def get_all_files_for_installed_packages_pip(packages: List[str]):
    """Placeholder docstring"""
    proc = subprocess.Popen(pipcmd + ["show", "-f"] + packages, stdout=subprocess.PIPE)
    with proc.stdout:
        lines = []
        for line in iter(proc.stdout.readline, b""):
            if line == b"---\n":
                yield lines
                lines = []
            else:
                lines.append(line)
    yield lines
    proc.wait(timeout=10)  # wait for the subprocess to exit


def get_all_files_for_installed_packages(packages: List[str]):
    """Placeholder docstring"""
    ret = {}
    for rawmsg in get_all_files_for_installed_packages_pip(packages):
        parser = email.parser.BytesParser(policy=email.policy.default)
        msg = parser.parsebytes(b"".join(iter(rawmsg)))
        if not msg.get("Files"):
            continue
        ret[msg.get("Name")] = {
            Path(msg.get("Location")).joinpath(x) for x in msg.get("Files").split()
        }

    return ret


def batched(iterable, n):
    """Batch data into tuples of length n. The last batch may be shorter."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while True:
        batch = tuple(itertools.islice(it, n))
        if not batch:
            break
        yield batch


def get_all_installed_packages():
    """Placeholder docstring"""
    proc = subprocess.run(pipcmd + ["list", "--format", "json"], stdout=subprocess.PIPE, check=True)
    return json.loads(proc.stdout)


def map_package_names_to_files(package_names: List[str]):
    """Placeholder docstring"""
    m = {}
    batch_size = 20
    with tqdm.tqdm(total=len(package_names), desc="Scanning for dependencies", ncols=100) as pbar:
        for pkg_names in batched(package_names, batch_size):
            m.update(get_all_files_for_installed_packages(list(pkg_names)))
            pbar.update(batch_size)
    return m


def get_currently_used_packages():
    """Placeholder docstring"""
    all_installed_packages = get_all_installed_packages()
    package_to_file_names = map_package_names_to_files([x["name"] for x in all_installed_packages])

    currently_used_files = {
        Path(m.__file__)
        for m in sys.modules.values()
        if inspect.ismodule(m) and hasattr(m, "__file__") and m.__file__
    }

    currently_used_packages = set()
    for file in currently_used_files:
        for package in package_to_file_names:
            if file in package_to_file_names[package]:
                currently_used_packages.add(package)
    return currently_used_packages


def get_requirements_for_pkl_file(pkl_path: Path, dest: Path):
    """Placeholder docstring"""
    with open(pkl_path, mode="rb") as file:
        cloudpickle.load(file)

    currently_used_packages = get_currently_used_packages()

    with open(dest, mode="w+") as out:
        for x in get_all_installed_packages():
            name = x["name"]
            version = x["version"]
            # skip only for dev
            if name == "boto3":
                boto3_version = boto3.__version__
                out.write(f"boto3=={boto3_version}\n")
            elif name in currently_used_packages:
                out.write(f"{name}=={version}\n")


def get_all_requirements(dest: Path):
    """Placeholder docstring"""
    all_installed_packages = get_all_installed_packages()

    with open(dest, mode="w+") as out:
        for package_info in all_installed_packages:
            name = package_info.get("name")
            version = package_info.get("version")

            out.write(f"{name}=={version}\n")



"""Module for SMD Server"""

from __future__ import absolute_import

import logging
import platform
from sagemaker.core.common_utils import _is_s3_uri
from sagemaker.core.helper.session_helper import Session
from sagemaker.core.s3.utils import determine_bucket_and_prefix, parse_s3_url
from sagemaker.core import fw_utils
from sagemaker.serve.utils.uploader import upload

logger = logging.getLogger(__name__)


class SageMakerSmdServer:
    """Placeholder docstring"""

    def _upload_smd_artifacts(
        self,
        model_path: str,
        sagemaker_session: Session,
        secret_key: str,
        s3_model_data_url: str = None,
        image: str = None,
        should_upload_artifacts: bool = False,
    ):
        """Tar the model artifact and upload to S3 bucket, then prepare for the environment variables"""
        s3_upload_path = None
        if _is_s3_uri(model_path):
            s3_upload_path = model_path
        elif should_upload_artifacts:
            if s3_model_data_url:
                bucket, key_prefix = parse_s3_url(url=s3_model_data_url)
            else:
                bucket, key_prefix = None, None

            code_key_prefix = fw_utils.model_code_key_prefix(key_prefix, None, image)

            bucket, code_key_prefix = determine_bucket_and_prefix(
                bucket=bucket, key_prefix=code_key_prefix, sagemaker_session=sagemaker_session
            )

            logger.debug(
                "Uploading the model resources to bucket=%s, key_prefix=%s.",
                bucket,
                code_key_prefix,
            )
            s3_upload_path = upload(sagemaker_session, model_path, bucket, code_key_prefix)
            logger.debug("Model resources uploaded to: %s", s3_upload_path)

        env_vars = {
            "SAGEMAKER_INFERENCE_CODE_DIRECTORY": "/opt/ml/model/code",
            "SAGEMAKER_INFERENCE_CODE": "inference.handler",
            "SAGEMAKER_REGION": sagemaker_session.boto_region_name,
            "SAGEMAKER_SERVE_SECRET_KEY": secret_key,
            "LOCAL_PYTHON": platform.python_version(),
        }
        return s3_upload_path, env_vars



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains logic for setting defaults in ModelBuilder."""
from __future__ import absolute_import

from typing import Optional, Dict, List, Union
from sagemaker.core.helper.pipeline_variable import PipelineVariable
from dataclasses import dataclass


@dataclass
class Network:
    """Network configuration for model deployment."""
    subnets: Optional[List[str]] = None
    security_group_ids: Optional[List[str]] = None
    enable_network_isolation: bool = False
    vpc_config: Optional[Dict[str, List[Union[str, PipelineVariable]]]] = None


@dataclass
class Compute:
    """Compute configuration for model deployment."""
    instance_type: Optional[str]
    instance_count: Optional[int] = 1


# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Holds the validation logic used for the .optimize() function. INTERNAL only"""
from __future__ import absolute_import

import textwrap
import logging
from typing import Any, Dict, Set, Optional
from enum import Enum
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class _OptimizationContainer(Enum):
    """Optimization containers"""

    TRT = "TRT"
    VLLM = "vLLM"
    NEURON = "Neuron"


class _OptimizationCombination(BaseModel):
    """Optimization ruleset data structure for comparing input to ruleset"""

    optimization_container: _OptimizationContainer = None
    compilation: Set[Optional[bool]]
    speculative_decoding: Set[Optional[bool]]
    sharding: Set[Optional[bool]]
    quantization_technique: Set[Optional[str]]

    def validate_against(self, optimization_combination, rule_set: _OptimizationContainer):
        """Validator for optimization containers"""

        # check the validity of each individual field
        if not optimization_combination.compilation.issubset(self.compilation):
            raise ValueError("Compilation")
        if not optimization_combination.quantization_technique.issubset(
            self.quantization_technique
        ):
            copy_quantization_technique = optimization_combination.quantization_technique.copy()
            raise ValueError(f"Quantization:{copy_quantization_technique.pop()}")
        if not optimization_combination.speculative_decoding.issubset(self.speculative_decoding):
            raise ValueError("Speculative Decoding")
        if not optimization_combination.sharding.issubset(self.sharding):
            raise ValueError("Sharding")

        # optimization technique combinations that need to be validated
        if optimization_combination.compilation and optimization_combination.speculative_decoding:
            is_compiled = optimization_combination.compilation.copy().pop()
            is_speculative_decoding = optimization_combination.speculative_decoding.copy().pop()
            if is_compiled and is_speculative_decoding:
                raise ValueError("Compilation and Speculative Decoding together")

        if rule_set == _OptimizationContainer.TRT:
            is_compiled = optimization_combination.compilation.copy().pop()
            is_quantized = optimization_combination.quantization_technique.copy().pop()
            if is_quantized and not is_compiled:
                raise ValueError(f"Quantization:{is_quantized} must be provided with Compilation")


TRUTHY_SET = {None, True}
FALSY_SET = {None, False}
TRT_CONFIGURATION = {
    "supported_instance_families": {"p4d", "p4de", "p5", "g5", "g6"},
    "optimization_combination": _OptimizationCombination(
        optimization_container=_OptimizationContainer.TRT,
        compilation=TRUTHY_SET,
        quantization_technique={None, "awq", "fp8", "smoothquant"},
        speculative_decoding=FALSY_SET,
        sharding=FALSY_SET,
    ),
}
VLLM_CONFIGURATION = {
    "supported_instance_families": {"p4d", "p4de", "p5", "g5", "g6"},
    "optimization_combination": _OptimizationCombination(
        optimization_container=_OptimizationContainer.VLLM,
        compilation=FALSY_SET,
        quantization_technique={None, "awq", "fp8"},
        speculative_decoding=TRUTHY_SET,
        sharding=TRUTHY_SET,
    ),
}
NEURON_CONFIGURATION = {
    "supported_instance_families": {"inf2", "trn1", "trn1n"},
    "optimization_combination": _OptimizationCombination(
        optimization_container=_OptimizationContainer.NEURON,
        compilation=TRUTHY_SET,
        quantization_technique={None},
        speculative_decoding=FALSY_SET,
        sharding=FALSY_SET,
    ),
}


def _validate_optimization_configuration(
    is_jumpstart: bool,
    instance_type: str,
    quantization_config: Dict[str, Any],
    compilation_config: Dict[str, Any],
    sharding_config: Dict[str, Any],
    speculative_decoding_config: Dict[str, Any],
):
    """Validate .optimize() input off of standard ruleset"""

    instance_family = None
    if instance_type:
        split_instance_type = instance_type.split(".")
        if len(split_instance_type) == 3:
            instance_family = split_instance_type[1]

    if (
        instance_family not in TRT_CONFIGURATION["supported_instance_families"]
        and instance_family not in VLLM_CONFIGURATION["supported_instance_families"]
        and instance_family not in NEURON_CONFIGURATION["supported_instance_families"]
    ):
        invalid_instance_type_msg = (
            f"Optimizations that uses {instance_type} instance type are "
            "not currently supported both on GPU and Neuron instances"
        )
        raise ValueError(invalid_instance_type_msg)

    quantization_technique = None
    if (
        quantization_config
        and quantization_config.get("OverrideEnvironment")
        and quantization_config.get("OverrideEnvironment").get("OPTION_QUANTIZE")
    ):
        quantization_technique = quantization_config.get("OverrideEnvironment").get(
            "OPTION_QUANTIZE"
        )

    optimization_combination = _OptimizationCombination(
        compilation={None if compilation_config is None else True},
        speculative_decoding={None if speculative_decoding_config is None else True},
        sharding={None if sharding_config is None else True},
        quantization_technique={quantization_technique},
    )

    # Check the case where no optimization combination is provided
    if (
        optimization_combination.compilation == {None}
        and optimization_combination.quantization_technique == {None}
        and optimization_combination.speculative_decoding == {None}
        and optimization_combination.sharding == {None}
    ):
        # JumpStart has defaults for Inf/Trn instances
        if is_jumpstart and instance_family in NEURON_CONFIGURATION["supported_instance_families"]:
            return
        raise ValueError(
            (
                "Optimizations that provide no optimization configs "
                "are currently not support on both GPU and Neuron instances."
            )
        )

    # Validate based off of instance type
    if instance_family in NEURON_CONFIGURATION["supported_instance_families"]:
        try:
            (
                NEURON_CONFIGURATION["optimization_combination"].validate_against(
                    optimization_combination, rule_set=_OptimizationContainer.NEURON
                )
            )
        except ValueError as neuron_compare_error:
            raise ValueError(
                (
                    f"Optimizations that use {neuron_compare_error} "
                    "are not supported on Neuron instances."
                )
            )
    else:
        if optimization_combination.compilation.copy().pop():  # Compilation is only enabled for TRT
            try:
                TRT_CONFIGURATION["optimization_combination"].validate_against(
                    optimization_combination, rule_set=_OptimizationContainer.TRT
                )
            except ValueError as trt_compare_error:
                raise ValueError(
                    (
                        f"Optimizations that use Compilation and {trt_compare_error} "
                        "are not supported for GPU instances."
                    )
                )
        else:
            try:
                (
                    VLLM_CONFIGURATION["optimization_combination"].validate_against(
                        optimization_combination, rule_set=_OptimizationContainer.VLLM
                    )
                )
            except ValueError as vllm_compare_error:
                try:  # try both VLLM and TRT to cover both rule sets
                    (
                        TRT_CONFIGURATION["optimization_combination"].validate_against(
                            optimization_combination, rule_set=_OptimizationContainer.TRT
                        )
                    )
                except ValueError as trt_compare_error:
                    if (
                        str(trt_compare_error)
                        == "Quantization:smoothquant must be provided with Compilation"
                    ):
                        raise ValueError(
                            f"Optimizations that use {trt_compare_error} for GPU instances."
                        )
                    if str(trt_compare_error) == str(vllm_compare_error):
                        raise ValueError(
                            (
                                f"Optimizations that use {trt_compare_error} "
                                "are not supported for GPU instances."
                            )
                        )
                    joint_error_msg = f"""
                    Optimization cannot be performed for the following reasons:
                    - Optimizations that use {trt_compare_error} are not supported for GPU instances.
                    - Optimizations that use {vllm_compare_error} are not supported for GPU instances.
                    """
                    raise ValueError(textwrap.dedent(joint_error_msg))



"""Module that defines the InProcessMode class"""

from __future__ import absolute_import

from pathlib import Path
import logging
from typing import Dict, Type, Optional
import time
from datetime import datetime, timedelta

from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.builder.schema_builder import SchemaBuilder
from sagemaker.serve.utils.exceptions import InProcessDeepPingException
from sagemaker.serve.model_server.in_process_model_server.in_process_server import InProcessServing
from sagemaker.core.helper.session_helper import Session

logger = logging.getLogger(__name__)

_PING_HEALTH_CHECK_FAIL_MSG = "Ping health check did not pass. Please review your inference code."


class InProcessMode(InProcessServing):
    """A class that holds methods to deploy model to a container in process environment"""

    def __init__(
        self,
        model: Optional[str],
        inference_spec: Optional[InferenceSpec],
        schema_builder: Type[SchemaBuilder],
        session: Session,
        model_path: str = None,
        env_vars: Dict = None,
    ):
        # pylint: disable=bad-super-call
        super().__init__()

        self.model = model
        self.inference_spec = inference_spec
        self.model_path = model_path
        self.env_vars = env_vars
        self.session = session
        self.schema_builder = schema_builder
        self._ping_local_server = None

    def load(self, model_path: str = None):
        """Loads model path, checks that path exists"""
        path = Path(model_path if model_path else self.model_path)
        if not path.exists():
            raise ValueError("model_path does not exist")
        if not path.is_dir():
            raise ValueError("model_path is not a valid directory")

        return self.inference_spec.load(str(path))

    def prepare(self):
        """Prepares the server"""

    def create_server(
        self,
        ping_fn,
    ):
        """Creating the fast api server and checking ping health."""

        logger.info("Waiting for fastapi server to start up...")

        logger.warning("Note: This is not a standard model server.")
        logger.warning("The model is being hosted directly on the FastAPI server.")

        self._ping_local_server = ping_fn
        self._start_serving()

        # allow some time for server to be ready.
        time.sleep(1)

        time_limit = datetime.now() + timedelta(seconds=5)
        healthy = True
        while True:
            final_pull = datetime.now() > time_limit
            if final_pull:
                break

            healthy, response = self._ping_local_server()

            if healthy:
                logger.debug("Ping health check has passed. Returned %s", str(response))
                break

        time.sleep(1)

        if not healthy:
            raise InProcessDeepPingException(_PING_HEALTH_CHECK_FAIL_MSG)

    def destroy_server(self):
        """Placeholder docstring"""
        self._stop_serving()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
Backward compatibility shim for sagemaker.serve.serverless.serverless_inference_config

This module has been moved to sagemaker.core.inference_config.
This file provides backward compatibility by re-exporting the class from its new location.

DEPRECATED: Import from sagemaker.core.inference_config instead.
"""
from __future__ import absolute_import

import warnings

# Import from new Core location
from sagemaker.core.inference_config import ServerlessInferenceConfig

# Emit deprecation warning when this module is imported
warnings.warn(
    "Importing from sagemaker.serve.serverless.serverless_inference_config is deprecated. "
    "The ServerlessInferenceConfig class has been moved to sagemaker.core.inference_config. "
    "Please update your imports to:\n"
    "  from sagemaker.core.inference_config import ServerlessInferenceConfig\n"
    "This compatibility shim will be removed in a future version.",
    DeprecationWarning,
    stacklevel=2
)

__all__ = ['ServerlessInferenceConfig']



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Holds the BedrockModelBuilder class."""
from __future__ import absolute_import

from typing import Optional, Dict, Any, Union

from sagemaker.core.helper.session_helper import Session
from sagemaker.core.resources import TrainingJob, ModelPackage

from sagemaker.train.model_trainer import ModelTrainer
from sagemaker.core.telemetry.telemetry_logging import _telemetry_emitter
from sagemaker.core.telemetry.constants import Feature


class BedrockModelBuilder:
    """Builder class for deploying models to Amazon Bedrock.
    
    This class provides functionality to deploy SageMaker models to Bedrock
    using either model import jobs or custom model creation, depending on
    the model type (Nova models vs. other models).

    Args:
        model: The model to deploy. Can be a ModelTrainer, TrainingJob, or ModelPackage instance.
    """

    def __init__(self, model: Optional[Union[ModelTrainer, TrainingJob, ModelPackage]]):
        """Initialize BedrockModelBuilder with a model instance.

        Args:
            model: The model to deploy. Can be a ModelTrainer, TrainingJob, or ModelPackage instance.
        """
        self.model = model
        self._bedrock_client = None
        self._sagemaker_client = None
        self.boto_session = Session().boto_session
        self.model_package = self._fetch_model_package() if model else None
        self.s3_model_artifacts = self._get_s3_artifacts() if model else None

    def _get_bedrock_client(self):
        """Get or create Bedrock client singleton.
        
        Returns:
            boto3.client: Bedrock client instance.
        """
        if self._bedrock_client is None:
            self._bedrock_client = self.boto_session.client("bedrock")
        return self._bedrock_client

    def _get_sagemaker_client(self):
        """Get or create SageMaker client singleton.
        
        Returns:
            boto3.client: SageMaker client instance.
        """
        if self._sagemaker_client is None:
            self._sagemaker_client = self.boto_session.client("sagemaker")
        return self._sagemaker_client

    @_telemetry_emitter(feature=Feature.MODEL_CUSTOMIZATION, func_name="BedrockModelBuilder.deploy")
    def deploy(
            self,
            job_name: Optional[str] = None,
            imported_model_name: Optional[str] = None,
            custom_model_name: Optional[str] = None,
            role_arn: Optional[str] = None,
            job_tags: Optional[list] = None,
            imported_model_tags: Optional[list] = None,
            model_tags: Optional[list] = None,
            client_request_token: Optional[str] = None,
            imported_model_kms_key_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Deploy the model to Bedrock.
        
        Automatically detects if the model is a Nova model and uses the appropriate
        Bedrock API (create_custom_model for Nova, create_model_import_job for others).

        Args:
            job_name: Name for the model import job (non-Nova models only).
            imported_model_name: Name for the imported model (non-Nova models only).
            custom_model_name: Name for the custom model (Nova models only).
            role_arn: IAM role ARN with permissions for Bedrock operations.
            job_tags: Tags for the import job (non-Nova models only).
            imported_model_tags: Tags for the imported model (non-Nova models only).
            model_tags: Tags for the custom model (Nova models only).
            client_request_token: Unique token for idempotency (non-Nova models only).
            imported_model_kms_key_id: KMS key ID for encryption (non-Nova models only).

        Returns:
            Response from Bedrock API containing job ARN or model ARN.
            
        Raises:
            ValueError: If required parameters are missing for the detected model type.
        """
        container = self.model_package.inference_specification.containers[0]
        is_nova = (hasattr(container, 'base_model') and container.base_model and
                   hasattr(container.base_model, 'recipe_name') and container.base_model.recipe_name and
                   "nova" in container.base_model.recipe_name.lower()) or \
                  (hasattr(container, 'base_model') and container.base_model and
                   hasattr(container.base_model, 'hub_content_name') and container.base_model.hub_content_name and
                   "nova" in container.base_model.hub_content_name.lower())

        if is_nova:
            params = {
                "modelName": custom_model_name,
                "modelSourceConfig": {"s3DataSource": {"s3Uri": self.s3_model_artifacts}},
                "roleArn": role_arn,
            }
            if model_tags:
                params["modelTags"] = model_tags
            params = {k: v for k, v in params.items() if v is not None}
            return self._get_bedrock_client().create_custom_model(**params)
        else:
            model_data_source = {"s3DataSource": {"s3Uri": self.s3_model_artifacts}}
            params = {
                "jobName": job_name,
                "importedModelName": imported_model_name,
                "roleArn": role_arn,
                "modelDataSource": model_data_source,
                "jobTags": job_tags,
                "importedModelTags": imported_model_tags,
                "clientRequestToken": client_request_token,
                "importedModelKmsKeyId": imported_model_kms_key_id,
            }
            params = {k: v for k, v in params.items() if v is not None}
            return self._get_bedrock_client().create_model_import_job(**params)

    def _fetch_model_package(self) -> Optional[ModelPackage]:
        """Fetch the ModelPackage from the provided model.
        
        Extracts ModelPackage from ModelTrainer, TrainingJob, or returns
        the ModelPackage directly if that's what was provided.
        
        Returns:
            ModelPackage instance or None if no model was provided.
        """
        if isinstance(self.model, ModelPackage):
            return self.model
        if isinstance(self.model, TrainingJob):
            return ModelPackage.get(self.model.output_model_package_arn)
        if isinstance(self.model, ModelTrainer):
            return ModelPackage.get(self.model._latest_training_job.output_model_package_arn)
        return None

    def _get_s3_artifacts(self) -> Optional[str]:
        """Extract S3 URI of model artifacts from the model package.
        
        For Nova models, fetches checkpoint URI from manifest.json in training job output.
        For other models, returns the model data source S3 URI.
        
        Returns:
            S3 URI string of the model artifacts, or None if not available.
        """
        if not self.model_package:
            return None
        
        container = self.model_package.inference_specification.containers[0]
        is_nova = (hasattr(container, 'base_model') and container.base_model and 
                  hasattr(container.base_model, 'recipe_name') and container.base_model.recipe_name and
                  "nova" in container.base_model.recipe_name.lower()) or \
                  (hasattr(container, 'base_model') and container.base_model and
                   hasattr(container.base_model, 'hub_content_name') and container.base_model.hub_content_name and
                   "nova" in container.base_model.hub_content_name.lower())
        
        if is_nova and isinstance(self.model, TrainingJob):
            return self._get_checkpoint_uri_from_manifest()
        
        if hasattr(container, 'model_data_source') and container.model_data_source:
            if hasattr(container.model_data_source, 's3_data_source') and container.model_data_source.s3_data_source:
                return container.model_data_source.s3_data_source.s3_uri
        return None
    
    def _get_checkpoint_uri_from_manifest(self) -> Optional[str]:
        """Get checkpoint URI from manifest.json for Nova models.
        
        Steps:
        1. Fetch S3 model artifacts from training job
        2. Go one level up in directory
        3. Find manifest.json
        4. Fetch checkpoint_s3_bucket from manifest
        
        Returns:
            Checkpoint URI from manifest.json.
            
        Raises:
            ValueError: If manifest.json cannot be found or parsed.
        """
        import json
        from urllib.parse import urlparse
        import logging
        
        logger = logging.getLogger(__name__)
        
        if not isinstance(self.model, TrainingJob):
            raise ValueError("Model must be a TrainingJob instance for Nova models")
        
        # Step 1: Get S3 model artifacts from training job
        s3_artifacts = self.model.model_artifacts.s3_model_artifacts
        if not s3_artifacts:
            raise ValueError("No S3 model artifacts found in training job")
        
        logger.info(f"S3 artifacts path: {s3_artifacts}")
        
        # Step 2: Construct manifest path (same directory as model artifacts)
        # s3://bucket/path/output/model.tar.gz -> s3://bucket/path/output/output/manifest.json
        parts = s3_artifacts.rstrip('/').rsplit('/', 1)
        manifest_path = parts[0] + '/output/manifest.json'
        
        logger.info(f"Manifest path: {manifest_path}")
        
        # Step 3: Find and read manifest.json
        parsed = urlparse(manifest_path)
        bucket = parsed.netloc
        manifest_key = parsed.path.lstrip('/')
        
        logger.info(f"Looking for manifest at s3://{bucket}/{manifest_key}")
        
        s3_client = self.boto_session.client('s3')
        try:
            response = s3_client.get_object(Bucket=bucket, Key=manifest_key)
            manifest = json.loads(response['Body'].read().decode('utf-8'))
            logger.info(f"Manifest content: {manifest}")
            
            # Step 4: Fetch checkpoint_s3_bucket from manifest
            checkpoint_uri = manifest.get('checkpoint_s3_bucket')
            if not checkpoint_uri:
                raise ValueError(f"'checkpoint_s3_bucket' not found in manifest. Available keys: {list(manifest.keys())}")
            
            logger.info(f"Checkpoint URI: {checkpoint_uri}")
            return checkpoint_uri
        except s3_client.exceptions.NoSuchKey:
            raise ValueError(f"manifest.json not found at s3://{bucket}/{manifest_key}")
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to parse manifest.json: {e}")
        except Exception as e:
            raise ValueError(f"Error reading manifest.json: {e}")


"""Script to scrape account IDs for 1p image URIs defined in src/sagemaker/image_uri_config."""

from __future__ import absolute_import
import os

import json

image_uri_config_dir_path = "../../image_uri_config"

account_ids = set()


def extract_account_ids(json_obj):
    """Traverses JSON object until account_ids are found under 'registries'."""
    if isinstance(json_obj, dict):
        for key, value in json_obj.items():
            if key == "registries" and isinstance(value, dict):
                account_ids.update(value.values())
            else:
                extract_account_ids(value)
    elif isinstance(json_obj, list):
        for item in json_obj:
            extract_account_ids(item)


for filename in os.listdir(image_uri_config_dir_path):
    with open(os.path.join(image_uri_config_dir_path, filename), "r") as f:
        try:
            data = json.load(f)
            for version in data["versions"]:
                extract_account_ids(data["versions"][version])
        except KeyError:
            # JSON objects in image_uri_config/ don't have consistent formatting.
            # Some include job types i.e. 'eia', 'inference', 'training', etc.
            # see tensorflow.json for an example.
            for job_type in data:
                for version in data[job_type]["versions"]:
                    extract_account_ids(data[job_type]["versions"][version])
        except json.JSONDecodeError:
            print(f"Error processing file: {filename}")

print(account_ids)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Utilites for identifying and analyzing local gpu hardware"""
from __future__ import absolute_import

import subprocess
import multiprocessing
import logging
import shutil
from math import ceil, floor
from platform import system
from pathlib import Path
import psutil

from sagemaker.core.helper.session_helper import Session

logger = logging.getLogger(__name__)

# key = vCPUs
# for A100, vCPUs are the same. Distinguish by GPU memory
hardware_lookup = {
    "NVIDIA H100": {192: "ml.p5.48xlarge"},
    "NVIDIA A100": {
        96: {
            40: "ml.p4d.24xlarge",
            80: "ml.p4de.24xlarge",
        }
    },
    "NVIDIA V100": {
        8: "ml.p3.2xlarge",
        32: "ml.p3.8xlarge",
        64: "ml.p3.16xlarge",
        96: "ml.p3dn.24xlarge",
    },
    "NVIDIA K80": {
        4: "ml.p2.xlarge",
        32: "ml.p2.8xlarge",
        64: "ml.p2.16xlarge",
    },
    "NVIDIA T4": {
        4: "ml.g4dn.xlarge",
        8: "ml.g4dn.2xlarge",
        16: "ml.g4dn.4xlarge",
        32: "ml.g4dn.8xlarge",
        48: "ml.g4dn.12xlarge",
        64: "ml.g4dn.16xlarge",
    },
    "NVIDIA A10G": {
        4: "ml.g5n.xlarge",
        8: "ml.g5.2xlarge",
        16: "ml.g5.4xlarge",
        32: "ml.g5.8xlarge",
        48: "ml.g5.12xlarge",
        64: "ml.g5.16xlarge",
        96: "ml.g5.24xlarge",
        192: "ml.g5.48xlarge",
    },
}


fallback_gpu_resource_mapping = {
    "ml.p5.48xlarge": 8,
    "ml.p4d.24xlarge": 8,
    "ml.p4de.24xlarge": 8,
    "ml.p3.2xlarge": 1,
    "ml.p3.8xlarge": 4,
    "ml.p3.16xlarge": 8,
    "ml.p3dn.24xlarge": 8,
    "ml.p2.xlarge": 1,
    "ml.p2.8xlarge": 8,
    "ml.p2.16xlarge": 16,
    "ml.g4dn.xlarge": 1,
    "ml.g4dn.2xlarge": 1,
    "ml.g4dn.4xlarge": 1,
    "ml.g4dn.8xlarge": 1,
    "ml.g4dn.16xlarge": 1,
    "ml.g4dn.12xlarge": 4,
    "ml.g5n.xlarge": 1,
    "ml.g5.2xlarge": 1,
    "ml.g5.4xlarge": 1,
    "ml.g5.8xlarge": 1,
    "ml.g5.16xlarge": 1,
    "ml.g5.12xlarge": 4,
    "ml.g5.24xlarge": 4,
    "ml.g5.48xlarge": 8,
}


def _get_available_gpus(log=True):
    """Detect the GPUs available on the device and their available resources"""
    try:
        gpu_query = ["nvidia-smi", "--query-gpu=name,memory.free", "--format=csv"]
        gpu_info_csv = subprocess.run(gpu_query, stdout=subprocess.PIPE, check=True)
        gpu_info = gpu_info_csv.stdout.decode("utf-8").splitlines()[1:]

        if log:
            logger.info("CUDA enabled hardware on the device: %s", gpu_info)
        return gpu_info
    except Exception as e:  # pylint: disable=W0703
        # for nvidia-smi to run, a cuda driver must be present
        logger.warning(
            "CUDA is not enabled on your device. %s. "
            "Please run ModelBuilder on CUDA enabled hardware "
            "to deploy locally.",
            str(e),
        )
        return None


def _get_nb_instance():
    """Placeholder docstring"""
    gpu_info = _get_available_gpus(False)
    if not gpu_info:
        return None

    gpu_name, gpu_mem = gpu_info[0].split(", ")
    cpu_count = multiprocessing.cpu_count()

    hardware = hardware_lookup.get(gpu_name)
    if not hardware:
        logger.info("Could not detect the instance type for GPU: %s, %s", gpu_name, gpu_mem)
        return None

    if gpu_name == "NVIDIA A100":
        gpu_mem_mib = int(gpu_mem.split(" ")[0])
        gpu_mem_gb = gpu_mem_mib * 1024**2 / 1000**3

        if ceil(gpu_mem_gb) in hardware:
            instance_type = hardware.get(int(ceil(gpu_mem_gb)))
        else:
            instance_type = hardware.get(int(floor(gpu_mem_gb)))
    else:
        instance_type = hardware.get(cpu_count)

    logger.info(
        "Local instance_type %s detected. "
        "%s will be default when deploying to a SageMaker Endpoint. "
        "This default can be overriden in model.deploy()",
        instance_type,
        instance_type,
    )
    return instance_type


def _get_ram_usage_mb():
    """Placeholder docstring"""
    return psutil.virtual_memory()[3] / 1000000


def _check_disk_space(model_path: str):
    """Placeholder docstring"""
    usage = shutil.disk_usage(model_path)
    percentage_used = usage[1] / usage[0]
    if percentage_used >= 0.5:
        logger.warning(
            "%s percent of disk space used. Please consider freeing up disk space "
            "or increasing the EBS volume if you are on a SageMaker Notebook.",
            percentage_used * 100,
        )


def _check_docker_disk_usage():
    """Fetch the local docker container disk usage.

    Args:
        None
    Returns:
        None
    """
    try:
        docker_path = "/var/lib/docker"
        # Windows, MacOS and Linux based docker installations work differently.
        if system().lower() == "windows":
            docker_path = str(Path.home()) + "C:\\ProgramData\\Docker"
        elif system().lower() == "darwin":
            docker_path = str(Path.home()) + "/Library/Containers/com.docker.docker/Data/vms/0/"
        elif system().lower() == "linux":
            docker_path = "/var/lib/docker"
        usage = shutil.disk_usage(docker_path)
        percentage_used = usage[1] / usage[0]
        if percentage_used >= 0.5:
            logger.warning(
                "%s percent of docker disk space at %s is used. "
                "Please consider freeing up disk space or increasing the EBS volume if you "
                "are on a SageMaker Notebook.",
                percentage_used * 100,
                docker_path,
            )
        else:
            logger.info(
                "%s percent of docker disk space at %s is used.", percentage_used * 100, docker_path
            )
    except Exception as e:  # pylint: disable=W0703
        logger.warning(
            "Unable to check docker volume utilization at the expected path %s. %s",
            docker_path,
            str(e),
        )


def _get_gpu_info(instance_type: str, session: Session) -> int:
    """Get GPU info for the provided instance"""
    ec2_client = session.boto_session.client("ec2")

    split_instance = instance_type.split(".")
    split_instance.pop(0)

    ec2_instance = ".".join(split_instance)

    instance_info = ec2_client.describe_instance_types(InstanceTypes=[ec2_instance])

    gpus_info = instance_info.get("InstanceTypes")[0].get("GpuInfo")

    if gpus_info:
        return gpus_info.get("Gpus")[0].get("Count")
    raise ValueError("Provided instance_type is not GPU enabled.")


def _get_gpu_info_fallback(instance_type: str) -> int:
    """Get GPU info for the provided instance fallback"""
    available_gpus = fallback_gpu_resource_mapping.get(instance_type)
    if not available_gpus:
        raise ValueError("Provided instance_type is not GPU enabled.")
    return available_gpus



"""Module that defines the LocalContainerMode class"""

from __future__ import absolute_import
from pathlib import Path
import logging
from datetime import datetime, timedelta
from typing import Dict, Type
import base64
import time
import subprocess
import docker

from sagemaker.serve.model_server.tensorflow_serving.server import LocalTensorflowServing
from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.builder.schema_builder import SchemaBuilder
from sagemaker.serve.utils.logging_agent import pull_logs
from sagemaker.serve.utils.types import ModelServer
from sagemaker.serve.utils.exceptions import LocalDeepPingException
from sagemaker.serve.model_server.torchserve.server import LocalTorchServe
from sagemaker.serve.model_server.djl_serving.server import LocalDJLServing
from sagemaker.serve.model_server.triton.server import LocalTritonServer
from sagemaker.serve.model_server.tgi.server import LocalTgiServing
from sagemaker.serve.model_server.tei.server import LocalTeiServing
from sagemaker.serve.model_server.multi_model_server.server import LocalMultiModelServer
from sagemaker.core.helper.session_helper import Session

logger = logging.getLogger(__name__)

_PING_HEALTH_CHECK_INTERVAL_SEC = 5

_PING_HEALTH_CHECK_FAIL_MSG = (
    "Container did not pass the ping health check. "
    + "Please increase container_timeout_seconds or review your inference code."
)


class LocalContainerMode(
    LocalTorchServe,
    LocalDJLServing,
    LocalTritonServer,
    LocalTgiServing,
    LocalMultiModelServer,
    LocalTensorflowServing,
):
    """A class that holds methods to deploy model to a container in local environment"""

    def __init__(
        self,
        model_server: ModelServer,
        inference_spec: Type[InferenceSpec],
        schema_builder: Type[SchemaBuilder],
        session: Session,
        model_path: str = None,
        env_vars: Dict = None,
    ):
        # pylint: disable=bad-super-call
        super().__init__()
        super(LocalDJLServing, self).__init__()
        super(LocalTritonServer, self).__init__()

        self.inference_spec = inference_spec
        self.model_path = model_path
        self.env_vars = env_vars
        self.session = session
        self.schema_builder = schema_builder
        self.ecr = session.boto_session.client("ecr")
        self.model_server = model_server
        self.client = None
        self.container = None
        self.secret_key = None
        self._ping_container = None
        self._invoke_serving = None

    def load(self, model_path: str = None):
        """Placeholder docstring"""
        path = Path(model_path if model_path else self.model_path)
        if not path.exists():
            raise Exception("model_path does not exist")
        if not path.is_dir():
            raise Exception("model_path is not a valid directory")

        return self.inference_spec.load(str(path))

    def prepare(self):
        """Placeholder docstring"""

    def create_server(
        self,
        image: str,
        container_timeout_seconds: int,
        secret_key: str,
        container_config: Dict,
        ping_fn = None,
        env_vars: Dict[str, str] = None,
        model_path: str = None,
        jumpstart: bool = False,
    ):
        """Placeholder docstring"""

        self._pull_image(image=image)

        self.destroy_server()

        logger.info("Waiting for model server %s to start up...", self.model_server)

        self._ping_container = ping_fn or self._ping_container
        if self.model_server == ModelServer.TRITON:
            self._start_triton_server(
                docker_client=self.client,
                model_path=model_path if model_path else self.model_path,
                image_uri=image,
                secret_key=secret_key,
                env_vars=env_vars if env_vars else self.env_vars,
            )
        elif self.model_server == ModelServer.DJL_SERVING:
            self._start_djl_serving(
                client=self.client,
                image=image,
                model_path=model_path if model_path else self.model_path,
                secret_key=secret_key,
                env_vars=env_vars if env_vars else self.env_vars,
            )
        elif self.model_server == ModelServer.TORCHSERVE:
            self._start_torch_serve(
                client=self.client,
                image=image,
                model_path=model_path if model_path else self.model_path,
                secret_key=secret_key,
                env_vars=env_vars if env_vars else self.env_vars,
            )
        elif self.model_server == ModelServer.TGI:
            self._start_tgi_serving(
                client=self.client,
                image=image,
                model_path=model_path if model_path else self.model_path,
                secret_key=secret_key,
                env_vars=env_vars if env_vars else self.env_vars,
                jumpstart=jumpstart,
            )
        elif self.model_server == ModelServer.MMS:
            self._start_serving(
                client=self.client,
                image=image,
                model_path=model_path if model_path else self.model_path,
                secret_key=secret_key,
                env_vars=env_vars if env_vars else self.env_vars,
            )
        elif self.model_server == ModelServer.TENSORFLOW_SERVING:
            self._start_tensorflow_serving(
                client=self.client,
                image=image,
                model_path=model_path if model_path else self.model_path,
                secret_key=secret_key,
                env_vars=env_vars if env_vars else self.env_vars,
            )
        elif self.model_server == ModelServer.TEI:
            tei_serving = LocalTeiServing()
            tei_serving._start_tei_serving(
                client=self.client,
                image=image,
                model_path=model_path if model_path else self.model_path,
                secret_key=secret_key,
                env_vars=env_vars if env_vars else self.env_vars,
            )
            tei_serving.schema_builder = self.schema_builder
            self.container = tei_serving.container
            self._invoke_serving = tei_serving._invoke_tei_serving

        # allow some time for container to be ready
        time.sleep(10)

        log_generator = self.container.logs(follow=True, stream=True)
        container_timeout_seconds = 1200
        time_limit = datetime.now() + timedelta(seconds=container_timeout_seconds)
        healthy = False
        while True:
            now = datetime.now()
            final_pull = now > time_limit
            pull_logs(
                (x.decode("UTF-8").rstrip() for x in log_generator),
                log_generator.close,
                datetime.now() + timedelta(seconds=_PING_HEALTH_CHECK_INTERVAL_SEC),
                now > time_limit,
            )

            if final_pull:
                break

            # allow some time for container to be ready
            time.sleep(10)
            healthy, response = self._ping_container()
            if healthy:
                logger.debug("Ping health check has passed. Returned %s", str(response))
                break

        if not healthy:
            raise LocalDeepPingException(_PING_HEALTH_CHECK_FAIL_MSG)

    def destroy_server(self):
        """Placeholder docstring"""
        if self.container:
            try:
                logger.debug("Stopping currently running container...")
                self.container.kill()
            except docker.errors.APIError as exc:
                if exc.response.status_code < 400 or exc.response.status_code > 499:
                    raise Exception("Error encountered when cleaning up local container") from exc
            self.container = None

    def _pull_image(self, image: str):
        """Pull image with proper error handling and early failure detection."""
        
        # Check if Docker is available first
        try:
            self.client = docker.from_env()
            self.client.ping()  # Test Docker connection
        except Exception as e:
            raise RuntimeError(
                f"Docker is not available or not running. Please ensure Docker is installed and running. "
                f"Error: {e}"
            ) from e
        
        # Handle ECR authentication for ECR images
        if self._is_ecr_image(image):
            try:
                encoded_token = (
                    self.ecr.get_authorization_token()
                    .get("authorizationData")[0]
                    .get("authorizationToken")
                )
                decoded_token = base64.b64decode(encoded_token).decode("utf-8")
                username, password = decoded_token.split(":")
                ecr_uri = image.split("/")[0]
                login_command = ["docker", "login", "-u", username, "-p", password, ecr_uri]
                
                result = subprocess.run(login_command, check=True, capture_output=True, text=True)
                logger.info("Successfully authenticated with ECR")
                
            except subprocess.CalledProcessError as e:
                error_msg = f"ECR authentication failed: {e.stderr if e.stderr else str(e)}"
                logger.error(error_msg)
                raise RuntimeError(error_msg) from e
            except Exception as e:
                error_msg = f"ECR authentication error: {str(e)}"
                logger.error(error_msg)
                raise RuntimeError(error_msg) from e
        
        # Pull the image
        try:
            logger.info("Pulling image %s from repository...", image)
            self.client.images.pull(image)
            logger.info("Successfully pulled image %s", image)
        except docker.errors.NotFound as e:
            raise ValueError(f"Could not find image '{image}' in repository") from e
        except docker.errors.APIError as e:
            raise RuntimeError(f"Failed to pull image '{image}': {e}") from e

    def _is_ecr_image(self, image: str) -> bool:
        """Check if image is from ECR."""
        return ".dkr.ecr." in image and ".amazonaws.com" in image




"""FastAPI requests"""

from __future__ import absolute_import

import asyncio
import io
import logging
import threading
import torch
from typing import Optional, Type

from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.builder.schema_builder import SchemaBuilder

logger = logging.getLogger(__name__)


try:
    import uvicorn
except ImportError:
    logger.error("Unable to import uvicorn, check if uvicorn is installed.")


try:
    from fastapi import FastAPI, Request, APIRouter
except ImportError:
    logger.error("Unable to import fastapi, check if fastapi is installed.")


class InProcessServer:
    """Generic In-Process Server for Serving Models using InferenceSpec"""

    def __init__(
        self,
        model: Optional[str] = None,
        inference_spec: Optional[InferenceSpec] = None,
        schema_builder: Type[SchemaBuilder] = None,
        task: Optional[str] = None,
    ):
        self._thread = None
        self._loop = None
        self._shutdown_event = threading.Event()
        self._router = APIRouter()
        self._task = task
        self.server = None
        self.port = None
        self.host = None
        self.model = model
        self.inference_spec = inference_spec
        self.schema_builder = schema_builder

        if self.inference_spec:
            # Use inference_spec to load the model
            self._load_model = self.inference_spec.load(model_dir=None)
        elif isinstance(self.model, str):
            try:
                # Use transformers pipeline to load the model
                try:
                    from transformers import pipeline, Pipeline
                except ImportError:
                    logger.error(
                        "Unable to import transformers, check if transformers is installed."
                    )

                device = 0 if torch.cuda.is_available() else -1

                self._load_model = pipeline(task, model=self.model, device=device)
            except Exception:
                logger.info("Falling back to SentenceTransformer for model loading.")
                try:
                    from sentence_transformers import SentenceTransformer
                except ImportError:
                    logger.error(
                        "Unable to import sentence-transformers, check if sentence-transformers is installed."
                    )

                self._load_model = SentenceTransformer(self.model)
        else:
            raise ValueError("Either inference_spec or model must be provided.")

        @self._router.post("/invoke")
        async def invoke(request: Request):
            """Generate text based on the provided prompt"""

            request_header = request.headers
            request_body = await request.body()
            content_type = request_header.get("Content-Type", None)
            input_data = schema_builder.input_deserializer.deserialize(
                io.BytesIO(request_body), content_type[0]
            )
            logger.debug(f"Received request: {input_data}")
            if self.inference_spec:
                response = self.inference_spec.invoke(input_data, self._load_model)
            else:
                input_data = input_data["inputs"] if "inputs" in input_data else input_data
                if isinstance(self._load_model, Pipeline):
                    response = self._load_model(input_data, max_length=30, num_return_sequences=1)
                else:
                    embeddings = self._load_model.encode(input_data, normalize_embeddings=True)
                    response = {"embeddings": embeddings.tolist()}
            return response

        self._create_server()

    def _create_server(self):
        """Placeholder docstring"""
        app = FastAPI()
        app.include_router(self._router)

        config = uvicorn.Config(
            app,
            host="127.0.0.1",
            port=9007,
            log_level="info",
            loop="asyncio",
            reload=True,
            use_colors=True,
        )

        self.server = uvicorn.Server(config)
        self.host = config.host
        self.port = config.port

    def start_server(self):
        """Starts the uvicorn server."""
        if not (self._thread and self._thread.is_alive()):
            logger.info("Waiting for a connection...")
            self._thread = threading.Thread(target=self._start_run_async_in_thread, daemon=True)
            self._thread.start()

    def stop_server(self):
        """Stops the Uvicorn server by setting the shutdown event."""
        if self._thread and self._thread.is_alive():
            logger.info("Shutting down the server...")
            self._shutdown_event.set()
            self.server.handle_exit(sig=0, frame=None)
            self._thread.join()

        logger.info("Server shutdown complete.")

    def _start_run_async_in_thread(self):
        """Placeholder docstring"""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self._serve())

    async def _serve(self):
        """Placeholder docstring"""
        await self.server.serve()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Holds the util functions used for MLflow model format"""
from __future__ import absolute_import

from pathlib import Path
from typing import Optional, Dict, Any, Union
import yaml
import logging
import shutil
import os

from sagemaker.core.helper.session_helper import Session
from sagemaker.core import image_uris
from sagemaker.serve.utils.types import ModelServer
from sagemaker.serve.detector.image_detector import _cast_to_compatible_version
from sagemaker.serve.model_format.mlflow.constants import (
    MLFLOW_FLAVOR_TO_PYTHON_PACKAGE_MAP,
    MLFLOW_PYFUNC,
    FLAVORS_WITH_FRAMEWORK_SPECIFIC_DLC_SUPPORT,
    DEFAULT_FW_USED_FOR_DEFAULT_IMAGE,
    DEFAULT_PYTORCH_VERSION,
    MLFLOW_METADATA_FILE,
    MLFLOW_PIP_DEPENDENCY_FILE,
    FLAVORS_DEFAULT_WITH_TF_SERVING,
    TENSORFLOW_SAVED_MODEL_NAME,
)

logger = logging.getLogger(__name__)


def _get_default_model_server_for_mlflow(deployment_flavor: str) -> ModelServer:
    """Get the default model server for mlflow based on deployment flavor.

    Args:
        deployment_flavor (str): The flavor mlflow model will be deployed with.

    Returns:
        str: The model server chosen for given model flavor.
    """
    if deployment_flavor in FLAVORS_DEFAULT_WITH_TF_SERVING:
        return ModelServer.TENSORFLOW_SERVING
    return ModelServer.TORCHSERVE


def _get_default_image_for_mlflow(python_version: str, region: str, instance_type: str) -> str:
    """Retrieves the default Docker image URI for MLflow deployments based on the specified Python

    version, AWS region, and instance type.

    Args:
        python_version (str): The Python version used in the environment,
            in the format 'major.minor.patch' (e.g., '3.8.6').
        region (str): The AWS region where the deployment is intended (e.g., 'us-east-1').
        instance_type (str): The type of AWS instance on which the deployment is intended.

    Returns:
        str: The URI of the default Docker image that matches the specified criteria.

    Raises:
        ValueError: If the function fails to retrieve a default Docker image URI based on the
            provided arguments.
    """
    major, minor, _ = python_version.split(".")
    shortened_py_version = f"py{major}{minor}"
    default_dlc = None
    # TODO: Dynamically getting fw version after beta
    try:
        default_dlc = image_uris.retrieve(
            framework=DEFAULT_FW_USED_FOR_DEFAULT_IMAGE,
            version=DEFAULT_PYTORCH_VERSION.get(shortened_py_version),
            region=region,
            image_scope="inference",
            py_version=shortened_py_version,
            instance_type=instance_type,
        )
    except ValueError:
        pass

    if default_dlc:
        logger.info(
            "Using default image %s based on inferred python version."
            " Proceeding with the the deployment.",
            default_dlc,
        )
        return default_dlc

    raise ValueError(
        f"Unable to find default image based on "
        f"python version {python_version} and region {region}"
    )


def _generate_mlflow_artifact_path(src_folder: str, artifact_name: str) -> str:
    """Generates the path to a specific MLflow model artifacts based on the source folder and

    convention.

    Args:
    - src_folder (str): The source folder where the MLflow model artifacts are stored.
    - artifact_name (str): The artifact name.

    Returns:
    - str: The path to the MLmodel file.
    """
    artifact_path = os.path.join(src_folder, artifact_name)

    if not os.path.isfile(artifact_path):
        raise FileNotFoundError(
            f"The {artifact_name} file does not exist in the specified "
            f"source folder: {src_folder}"
        )
    return artifact_path


def _get_all_flavor_metadata(mlmodel_path: str) -> Optional[Dict[str, Any]]:
    """Validates whether an MLflow MLmodel file is in YAML format and parses the supported flavors.

    Args:
    - mlmodel_path (str): Path to the MLmodel file.

    Returns:
    - Optional[List[str]]: A list of mlflow flavors model was saved with.

    Raises (ValueError): raises ValueError when any of the following scenario happen:
       1. file does not exist
       2. file cannot be parsed as YAML format.
       3. file does not have flavors as key.
    """
    if not os.path.isfile(mlmodel_path):
        raise ValueError(f"File does not exist: {mlmodel_path}")

    try:
        with open(mlmodel_path, "r") as file:
            mlmodel_content = yaml.safe_load(file)

            if "flavors" in mlmodel_content:
                # Extract and return the flavors as a list of keys
                return mlmodel_content["flavors"]
            else:
                raise ValueError("The 'flavors' key is missing in the MLmodel file.")
    except yaml.YAMLError as e:
        raise ValueError(f"Error parsing the file as YAML: {e}")


def _get_framework_version_from_requirements(flavor: str, requirements_path: str) -> Optional[str]:
    """Checks the version of a given MLflow flavor from a requirements.txt file.

    Args:
    - flavor (str): The MLflow flavor to look for (e.g., "tensorflow", "scikit-learn").
    - requirements_path (str): Path to the requirements.txt file.

    Returns:
    - Optional[str]: The version of the framework if found, None otherwise.

    Raises:
    - ValueError: If the requirements.txt file is not found.
    """

    python_package = MLFLOW_FLAVOR_TO_PYTHON_PACKAGE_MAP.get(flavor)
    if python_package is not None:
        try:
            with open(requirements_path, "r") as file:
                for line in file:
                    if line.startswith(python_package):
                        for separator in ("==", ">=", "<="):
                            if separator in line:
                                version = line.split(separator)[1].strip()
                                return version
                        logger.warning(f"Dependency of flavor {flavor} is not found.")
                        return None
        except FileNotFoundError:
            raise ValueError(f"File not found: {requirements_path}")

    logger.warning(f"Dependency of flavor {flavor} is not found.")
    return None


def _get_deployment_flavor(flavor_metadata: Optional[Dict[str, Any]]) -> str:
    """Get deployment flavor from all parsed flavor metadata.

    Args:
    - flavor_metadata (Optional[Dict[str, Any]]): Dictionary contains mlflow flavors and
        their metadata.

    Returns:
    - str: The flavor mlflow model is saved with. Default to pyfunc if no other flavor is found.
    """

    if not flavor_metadata:
        raise ValueError("Flavor metadata is not found")

    deployment_flavor = MLFLOW_PYFUNC
    for flavor in flavor_metadata:
        if flavor != MLFLOW_PYFUNC:
            deployment_flavor = flavor
            return deployment_flavor
    return deployment_flavor


def _get_python_version_from_parsed_mlflow_model_file(
    parsed_metadata: Dict[str, Any]
) -> Optional[str]:
    """Checks the python version of a given parsed MLflow model file.

    Args:
    - parsed_metadata (Dict[str, Any]): The parsed metadata of a MLflow model.

    Returns:
    - Optional[str]: Python version

    Raises:
    - ValueError: If python_function flavor is not found in metadata.
    """
    pyfunc_metadata = parsed_metadata.get(MLFLOW_PYFUNC)
    if pyfunc_metadata is not None:
        return pyfunc_metadata.get("python_version")
    raise ValueError(f"{MLFLOW_PYFUNC} cannot be found in MLmodel file.")


def _download_s3_artifacts(s3_path: str, dst_path: str, session: Session) -> None:
    """Downloads all artifacts from a specified S3 path to a local destination path.

    Args:
    - s3_path (str): The S3 path to download artifacts from (format: s3://bucket/key).
    - dst_path (str): The local file system path where artifacts should be downloaded.
    - session (Session): A boto3 Session object with configuration.
    """
    if not s3_path.startswith("s3://"):
        raise ValueError("Invalid S3 path provided. Must start with 's3://'")

    s3_bucket, s3_key = s3_path.replace("s3://", "").split("/", 1)

    s3 = session.boto_session.client("s3")

    os.makedirs(dst_path, exist_ok=True)

    paginator = s3.get_paginator("list_objects_v2")
    for page in paginator.paginate(Bucket=s3_bucket, Prefix=s3_key):
        for obj in page.get("Contents", []):
            key = obj["Key"]
            rel_path = os.path.relpath(key, s3_key)
            local_file_path = os.path.join(dst_path, rel_path)

            if not key.endswith("/"):
                local_file_dir = os.path.dirname(local_file_path)
                os.makedirs(local_file_dir, exist_ok=True)

                # Download the file
                logger.info(f"Downloading {key} to {local_file_path}")
                s3.download_file(s3_bucket, key, local_file_path)


def _copy_directory_contents(src_dir, dest_dir) -> None:
    """Copy everything under src_dir to dest_dir preserving the directory structure.

    Args:
        src_dir (str): The source directory path from which to copy the contents.
        dest_dir (str): The destination directory path where the contents of src_dir will be copied.
    """
    src_dir = os.path.normpath(src_dir)
    dest_dir = os.path.normpath(dest_dir)

    if src_dir == dest_dir:
        logger.info("Source and destination directories are the same. No action taken.")
        return

    for root, dirs, files in os.walk(src_dir):
        relative_path = os.path.relpath(root, src_dir)
        dest_path = os.path.join(dest_dir, relative_path)
        normalized_dest_path = os.path.normpath(dest_path)
        os.makedirs(normalized_dest_path, exist_ok=True)

        for file in files:
            file_src_path = os.path.join(root, file)
            file_dest_path = os.path.join(dest_path, file)
            normalized_file_dest_path = os.path.normpath(file_dest_path)
            shutil.copy2(file_src_path, normalized_file_dest_path)


def _select_container_for_mlflow_model(
    mlflow_model_src_path: str,
    deployment_flavor: str,
    region: str,
    instance_type: str,
) -> str:
    """Select framework specific DLC for mlflow model based on flavor, region and isntance type.

    Args:
        - mlflow_model_src_path (str): The local path to mlflow model artifacts.
        - deployment_flavor (str): The flavor mlflow model will be deployed with.
        - region (str): Current region.
        - instance_type (str): The type of AWS instance on which the deployment is intended.

    Returns:
        str: The image uri chosen for mlflow model.
    """
    # TODO: Extend this for model server specific DLC as well, ie. DJL, Triton
    requirement_path = _generate_mlflow_artifact_path(
        mlflow_model_src_path, MLFLOW_PIP_DEPENDENCY_FILE
    )
    mlflow_model_metadata_path = _generate_mlflow_artifact_path(
        mlflow_model_src_path, MLFLOW_METADATA_FILE
    )
    flavor_metadata = _get_all_flavor_metadata(mlflow_model_metadata_path)
    python_version = _get_python_version_from_parsed_mlflow_model_file(flavor_metadata)
    major_python_version, minor_python_version, _ = python_version.split(".")
    casted_python_version = (
        f"py{major_python_version}{minor_python_version}"
        if deployment_flavor == "pytorch"
        else f"py{major_python_version}"
    )

    if deployment_flavor not in FLAVORS_WITH_FRAMEWORK_SPECIFIC_DLC_SUPPORT:
        logger.warning(
            f"{deployment_flavor} flavor currently doesn't have optimized framework "
            f"specific DLC support. Defaulting to generic image..."
        )
        return _get_default_image_for_mlflow(python_version, region, instance_type)

    framework_to_use = FLAVORS_WITH_FRAMEWORK_SPECIFIC_DLC_SUPPORT.get(deployment_flavor)
    framework_version = _get_framework_version_from_requirements(framework_to_use, requirement_path)

    logger.info("Auto-detected deployment flavor is %s", deployment_flavor)
    logger.info("Auto-detected framework to use is %s", framework_to_use)
    logger.info("Auto-detected framework version is %s", framework_version)

    if framework_version is None:
        raise ValueError(
            (
                "Unable to auto detect framework version. Please provide framework %s as part of the "
                "requirements.txt file for deployment flavor %s"
            )
            % (framework_to_use, deployment_flavor)
        )

    casted_versions = (
        _cast_to_compatible_version(framework_to_use, framework_version)
        if framework_version
        else (None,)
    )

    image_uri = None
    for casted_version in filter(None, casted_versions):
        try:
            image_uri = image_uris.retrieve(
                framework=framework_to_use,
                region=region,
                version=casted_version,
                image_scope="inference",
                py_version=casted_python_version,
                instance_type=instance_type,
            )
            break
        except ValueError:
            pass

    if image_uri:
        logger.info("Auto detected %s. Proceeding with the the deployment.", image_uri)
        if deployment_flavor != "pytorch" and minor_python_version != "8":
            logger.warning(
                f"Image {image_uri} uses python version 3.8. Your python version used "
                f"for model is {python_version}. It is recommended to use the same "
                f"python version to avoid incompatibility."
            )
        return image_uri

    raise ValueError(
        (
            "Unable to auto detect a DLC for framework %s, framework version %s "
            "and python version %s. "
            "Please manually provide image_uri to ModelBuilder()"
        )
        % (deployment_flavor, framework_version, f"py{major_python_version}{minor_python_version}")
    )


def _validate_input_for_mlflow(model_server: ModelServer, deployment_flavor: str) -> None:
    """Validates arguments provided with mlflow models.

    Args:
        - model_server (ModelServer): Model server used for orchestrating mlflow model.
        - deployment_flavor (str): The flavor mlflow model will be deployed with.

    Raises:
    - ValueError: If model server is not torchserve.
    """
    if model_server != ModelServer.TORCHSERVE and model_server != ModelServer.TENSORFLOW_SERVING:
        raise ValueError(
            f"{model_server} is currently not supported for MLflow Model. "
            f"Please choose another model server."
        )
    if (
        model_server == ModelServer.TENSORFLOW_SERVING
        and deployment_flavor not in FLAVORS_DEFAULT_WITH_TF_SERVING
    ):
        raise ValueError(
            "Tensorflow Serving is currently only supported for the following "
            "deployment flavors: {}".format(FLAVORS_DEFAULT_WITH_TF_SERVING)
        )


def _get_saved_model_path_for_tensorflow_and_keras_flavor(model_path: str) -> Optional[str]:
    """Recursively searches for tensorflow saved model.

    Args:
        model_path (str): The root directory to start the search from.

    Returns:
        Optional[str]: The absolute path to the directory containing 'saved_model.pb'.
    """
    for dirpath, dirnames, filenames in os.walk(model_path):
        if TENSORFLOW_SAVED_MODEL_NAME in filenames:
            return os.path.abspath(dirpath)

    return None


def _move_contents(src_dir: Union[str, Path], dest_dir: Union[str, Path]) -> None:
    """Moves all contents of a source directory to a specified destination directory.

    Args:
        src_dir (Union[str, Path]): The path to the source directory.
        dest_dir (Union[str, Path]): The path to the destination directory.

    """
    _src_dir = Path(os.path.normpath(src_dir))
    _dest_dir = Path(os.path.normpath(dest_dir))

    _dest_dir.mkdir(parents=True, exist_ok=True)

    for item in _src_dir.iterdir():
        _dest_path = _dest_dir / item.name
        shutil.move(str(item), str(_dest_path))
    
    _src_dir.rmdir()



"""Module for Local TEI Serving"""

from __future__ import absolute_import

import requests
import logging
from pathlib import Path
from docker.types import DeviceRequest
from sagemaker.core.helper.session_helper import Session
from sagemaker.core import fw_utils
from sagemaker.core.s3.utils import determine_bucket_and_prefix, parse_s3_url, s3_path_join
from sagemaker.core.s3 import S3Uploader
from sagemaker.core.local.local_session import get_docker_host
from sagemaker.core.common_utils import _is_s3_uri

MODE_DIR_BINDING = "/opt/ml/model/"
_SHM_SIZE = "2G"
_DEFAULT_ENV_VARS = {
    "HF_HOME": "/opt/ml/model/",
    "HUGGINGFACE_HUB_CACHE": "/opt/ml/model/",
}

logger = logging.getLogger(__name__)


class LocalTeiServing:
    """LocalTeiServing class"""

    def _start_tei_serving(
        self, client: object, image: str, model_path: str, secret_key: str, env_vars: dict
    ):
        """Starts a local tei serving container.

        Args:
            client: Docker client
            image: Image to use
            model_path: Path to the model
            secret_key: Secret key to use for authentication
            env_vars: Environment variables to set
        """
        if env_vars and secret_key:
            env_vars["SAGEMAKER_SERVE_SECRET_KEY"] = secret_key

        self.container = client.containers.run(
            image,
            shm_size=_SHM_SIZE,
            device_requests=[DeviceRequest(count=-1, capabilities=[["gpu"]])],
            network_mode="host",
            detach=True,
            auto_remove=True,
            volumes={
                Path(model_path).joinpath("code"): {
                    "bind": MODE_DIR_BINDING,
                    "mode": "rw",
                },
            },
            environment=_update_env_vars(env_vars),
        )

    def _invoke_tei_serving(self, request: object, content_type: str, accept: str):
        """Invokes a local tei serving container.

        Args:
            request: Request to send
            content_type: Content type to use
            accept: Accept to use
        """
        try:
            response = requests.post(
                f"http://{get_docker_host()}:8080/invocations",
                data=request,
                headers={"Content-Type": content_type, "Accept": accept},
                timeout=600,
            )
            response.raise_for_status()
            return response.content
        except Exception as e:
            raise Exception("Unable to send request to the local container server") from e


class SageMakerTeiServing:
    """SageMakerTeiServing class"""

    def _upload_tei_artifacts(
        self,
        model_path: str,
        sagemaker_session: Session,
        s3_model_data_url: str = None,
        image: str = None,
        env_vars: dict = None,
        should_upload_artifacts: bool = False,
    ):
        """Uploads the model artifacts to S3.

        Args:
            model_path: Path to the model
            sagemaker_session: SageMaker session
            s3_model_data_url: S3 model data URL
            image: Image to use
            env_vars: Environment variables to set
            model_data_s3_path: S3 path to model data
            should_upload_artifacts: Whether to upload artifacts
        """
        model_data_url = None
        if _is_s3_uri(model_path):
            model_data_url = model_path
        elif should_upload_artifacts:
            if s3_model_data_url:
                bucket, key_prefix = parse_s3_url(url=s3_model_data_url)
            else:
                bucket, key_prefix = None, None

            code_key_prefix = fw_utils.model_code_key_prefix(key_prefix, None, image)

            bucket, code_key_prefix = determine_bucket_and_prefix(
                bucket=bucket, key_prefix=code_key_prefix, sagemaker_session=sagemaker_session
            )

            code_dir = Path(model_path).joinpath("code")

            s3_location = s3_path_join("s3://", bucket, code_key_prefix, "code")

            logger.debug("Uploading TEI Model Resources uncompressed to: %s", s3_location)

            model_data_url = S3Uploader.upload(
                str(code_dir),
                s3_location,
                None,
                sagemaker_session,
            )

        model_data = (
            {
                "S3DataSource": {
                    "CompressionType": "None",
                    "S3DataType": "S3Prefix",
                    "S3Uri": model_data_url + "/",
                }
            }
            if model_data_url
            else None
        )

        return (model_data, _update_env_vars(env_vars))


def _update_env_vars(env_vars: dict) -> dict:
    """Placeholder docstring"""
    updated_env_vars = {}
    updated_env_vars.update(_DEFAULT_ENV_VARS)
    if env_vars:
        updated_env_vars.update(env_vars)
    return updated_env_vars



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Holds constants used for lineage support"""
from __future__ import absolute_import


LINEAGE_POLLER_INTERVAL_SECS = 15
LINEAGE_POLLER_MAX_TIMEOUT_SECS = 120
TRACKING_SERVER_ARN_REGEX = r"arn:(.*?):sagemaker:(.*?):(.*?):mlflow-tracking-server/(.*?)$"
TRACKING_SERVER_CREATION_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
MODEL_BUILDER_MLFLOW_MODEL_PATH_LINEAGE_ARTIFACT_TYPE = "ModelBuilderInputModelData"
MLFLOW_S3_PATH = "S3"
MLFLOW_MODEL_PACKAGE_PATH = "ModelPackage"
MLFLOW_RUN_ID = "MLflowRunId"
MLFLOW_LOCAL_PATH = "Local"
MLFLOW_REGISTRY_PATH = "MLflowRegistry"
ERROR = "Error"
CODE = "Code"
CONTRIBUTED_TO = "ContributedTo"
VALIDATION_EXCEPTION = "ValidationException"



"""Exceptions used across different model builder invocations"""

from __future__ import absolute_import


class ModelBuilderException(Exception):
    """The base exception class for ModelBuilder exceptions."""

    fmt = "An unspecified error occurred"

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


class LocalDeepPingException(ModelBuilderException):
    """Raise when local model serving does not pass the deep ping check"""

    fmt = "Error Message: {message}"
    model_builder_error_code = 1

    def __init__(self, message):
        super().__init__(message=message)


class InProcessDeepPingException(ModelBuilderException):
    """Raise when in process model serving does not pass the deep ping check"""

    fmt = "Error Message: {message}"
    model_builder_error_code = 1

    def __init__(self, message):
        super().__init__(message=message)


class LocalModelOutOfMemoryException(ModelBuilderException):
    """Raise when local model serving fails to load the model"""

    fmt = "Error Message: {message}"
    model_builder_error_code = 2

    def __init__(self, message):
        super().__init__(message=message)


class LocalModelLoadException(ModelBuilderException):
    """Raise when local model serving fails to load the model"""

    fmt = "Error Message: {message}"
    model_builder_error_code = 3

    def __init__(self, message):
        super().__init__(message=message)


class LocalModelInvocationException(ModelBuilderException):
    """Raise when local model serving fails to invoke the model"""

    fmt = "Error Message: {message}"
    model_builder_error_code = 4

    def __init__(self, message):
        super().__init__(message=message)


class SkipTuningComboException(ModelBuilderException):
    """Raise when tuning combination should be admissible but is not"""

    fmt = "Error Message: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class TaskNotFoundException(ModelBuilderException):
    """Raise when HuggingFace task could not be found"""

    fmt = "Error Message: {message}"

    def __init__(self, message):
        super().__init__(message=message)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""
Local SageMaker Serve development package.

This __init__.py file imports key modules used by inference scripts to prevent
Python module resolution conflicts with external serve.py files.

The imports below "prime" the module cache so that sagemaker.serve is recognized
as a package, preventing conflicts when inference scripts import from submodules.
"""

from __future__ import absolute_import

# Strategic imports to prime module cache and prevent serve.py conflicts
# Match V2's imports to ensure same priming behavior
from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.utils.types import ModelServer
from sagemaker.serve.model_builder import ModelBuilder

__all__ = ["InferenceSpec", "ModelServer", "ModelBuilder"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Inference Recommender mixin for SageMaker model optimization.

This module provides the _InferenceRecommenderMixin class that enables SageMaker models
to use Inference Recommender for right-sizing and optimization recommendations.

Key Features:
- Automatic instance type and configuration recommendations
- Load testing with custom traffic patterns
- Performance optimization based on latency and throughput requirements
- Support for both Default and Advanced recommendation jobs

Example:
    Basic usage with a ModelBuilder::
    
        model_builder = ModelBuilder(model="my-model")
        model = model_builder.build()
        
        # Get right-sizing recommendations
        model.right_size(
            sample_payload_url="s3://my-bucket/sample-payload.json",
            supported_content_types=["application/json"],
            supported_instance_types=["ml.m5.large", "ml.m5.xlarge"]
        )
        
        # Deploy with recommendations
        predictor = model.deploy()
"""
from __future__ import absolute_import

# Standard library imports
import logging
import re
from typing import Any, Dict, List, Optional, Tuple

# SageMaker imports
from sagemaker.core.parameter import CategoricalParameter

# ========================================
# Constants
# ========================================

INFERENCE_RECOMMENDER_FRAMEWORK_MAPPING = {
    "xgboost": "XGBOOST",
    "sklearn": "SAGEMAKER-SCIKIT-LEARN", 
    "pytorch": "PYTORCH",
    "tensorflow": "TENSORFLOW",
    "mxnet": "MXNET",
}

# Setting LOGGER for backward compatibility, in case users import it
logger = LOGGER = logging.getLogger("sagemaker")


class Phase:
    """Traffic pattern phase configuration for Advanced Inference Recommendations.

    Defines a phase of load testing with specific duration, user count, and spawn rate.
    Multiple phases can be combined to create complex traffic patterns.
    
    Args:
        duration_in_seconds: How long this phase should run
        initial_number_of_users: Number of concurrent users at start of phase
        spawn_rate: Rate at which new users are added (users per second)
        
    Example:
        Create a ramp-up phase::
        
            phase = Phase(
                duration_in_seconds=300,  # 5 minutes
                initial_number_of_users=1,
                spawn_rate=2  # Add 2 users per second
            )
    """

    def __init__(self, duration_in_seconds: int, initial_number_of_users: int, spawn_rate: int) -> None:
        """Initialize a Phase for load testing.
        
        Args:
            duration_in_seconds: Duration of this phase in seconds
            initial_number_of_users: Starting number of concurrent users
            spawn_rate: Rate of adding new users (users per second)
        """
        self.to_json = {
            "DurationInSeconds": duration_in_seconds,
            "InitialNumberOfUsers": initial_number_of_users,
            "SpawnRate": spawn_rate,
        }


class ModelLatencyThreshold:
    """Latency threshold configuration for Advanced Inference Recommendations.

    Defines acceptable response latency limits for model inference.
    Used to filter recommendations based on performance requirements.
    
    Args:
        percentile: Latency percentile to measure (e.g., "P95", "P99")
        value_in_milliseconds: Maximum acceptable latency in milliseconds
        
    Example:
        Set P95 latency threshold::
        
            threshold = ModelLatencyThreshold(
                percentile="P95",
                value_in_milliseconds=100  # 100ms max P95 latency
            )
    """

    def __init__(self, percentile: str, value_in_milliseconds: int) -> None:
        """Initialize a ModelLatencyThreshold.
        
        Args:
            percentile: Latency percentile (e.g., "P95", "P99")
            value_in_milliseconds: Maximum latency threshold in milliseconds
        """
        self.to_json = {"Percentile": percentile, "ValueInMilliseconds": value_in_milliseconds}


class _InferenceRecommenderMixin:
    """Mixin class providing SageMaker Inference Recommender functionality.
    
    This mixin adds right-sizing capabilities to SageMaker models, enabling
    automatic instance type and configuration recommendations based on model
    performance requirements.
    
    The mixin provides:
    - Automatic framework detection from container images
    - Default and Advanced recommendation job types
    - Load testing with custom traffic patterns
    - Performance-based filtering and optimization
    
    This class is designed to be mixed into Model classes that have:
    - sagemaker_session: SageMaker session for API calls
    - role_arn: IAM role for job execution
    - model_name: Name of the model
    - image_uri: Container image URI (optional, for framework detection)
    """

    def right_size(
        self,
        sample_payload_url: Optional[str] = None,
        supported_content_types: Optional[List[str]] = None,
        supported_instance_types: Optional[List[str]] = None,
        job_name: Optional[str] = None,
        framework: Optional[str] = None,
        framework_version: Optional[str] = None,
        job_duration_in_seconds: Optional[int] = None,
        hyperparameter_ranges: Optional[List[Dict[str, CategoricalParameter]]] = None,
        phases: Optional[List[Phase]] = None,
        traffic_type: Optional[str] = None,
        max_invocations: Optional[int] = None,
        model_latency_thresholds: Optional[List[ModelLatencyThreshold]] = None,
        max_tests: Optional[int] = None,
        max_parallel_tests: Optional[int] = None,
        log_level: Optional[str] = "Verbose",
    ) -> "_InferenceRecommenderMixin":
        """Recommends an instance type for a SageMaker or BYOC model.

        Create a SageMaker ``Model`` or use a registered ``ModelPackage``,
        to start an Inference Recommender job.

        The name of the created model is accessible in the ``name`` field of
        this ``Model`` after right_size returns.

        Args:
            sample_payload_url (str): The S3 path where the sample payload is stored.
            supported_content_types: (list[str]): The supported MIME types for the input data.
            supported_instance_types (list[str]): A list of the instance types that this model
                is expected to work on. (default: None).
            job_name (str): The name of the Inference Recommendations Job. (default: None).
            framework (str): The machine learning framework of the Image URI.
                Only required to specify if you bring your own custom containers (default: None).
            job_duration_in_seconds (int): The maximum job duration that a job can run for.
                (default: None).
            hyperparameter_ranges (list[Dict[str, sagemaker.parameter.CategoricalParameter]]):
                Specifies the hyper parameters to be used during endpoint load tests.
                `instance_type` must be specified as a hyperparameter range.
                `env_vars` can be specified as an optional hyperparameter range. (default: None).
                Example::

                    hyperparameter_ranges = [{
                        'instance_types': CategoricalParameter(['ml.c5.xlarge', 'ml.c5.2xlarge']),
                        'OMP_NUM_THREADS': CategoricalParameter(['1', '2', '3', '4'])
                    }]

            phases (list[Phase]): Shape of the traffic pattern to use in the load test
                (default: None).
            traffic_type (str): Specifies the traffic pattern type. Currently only supports
                one type 'PHASES' (default: None).
            max_invocations (str): defines the minimum invocations per minute for the endpoint
                to support (default: None).
            model_latency_thresholds (list[ModelLatencyThreshold]): defines the maximum response
                latency for endpoints to support (default: None).
            max_tests (int): restricts how many endpoints in total are allowed to be
                spun up for this job (default: None).
            max_parallel_tests (int): restricts how many concurrent endpoints
                this job is allowed to spin up (default: None).
            log_level (str): specifies the inline output when waiting for right_size to complete
                (default: "Verbose").

        Returns:
            sagemaker.model.Model: A SageMaker ``Model`` object. See
            :func:`~sagemaker.model.Model` for full details.
        """
        # Auto-detect framework from image URI if not provided
        if not framework and hasattr(self, 'image_uri'):
            detected_framework, detected_version = self._extract_framework_from_image_uri()
            if detected_framework:
                # Convert framework enum to string if needed
                framework_str = getattr(detected_framework, 'value', str(detected_framework)).lower()
                framework = INFERENCE_RECOMMENDER_FRAMEWORK_MAPPING.get(
                    framework_str, 
                    str(detected_framework)
                )
            framework_version = framework_version or detected_version

        endpoint_configurations = self._convert_to_endpoint_configurations_json(
            hyperparameter_ranges=hyperparameter_ranges
        )
        traffic_pattern = self._convert_to_traffic_pattern_json(
            traffic_type=traffic_type, phases=phases
        )
        stopping_conditions = self._convert_to_stopping_conditions_json(
            max_invocations=max_invocations, model_latency_thresholds=model_latency_thresholds
        )
        resource_limit = self._convert_to_resource_limit_json(
            max_tests=max_tests, max_parallel_tests=max_parallel_tests
        )

        # Determine job type based on advanced parameters
        if endpoint_configurations or traffic_pattern or stopping_conditions or resource_limit:
            logger.info("Advanced job parameters specified. Running Advanced recommendation job...")
            job_type = "Advanced"
        else:
            logger.info("No advanced parameters specified. Running Default recommendation job...")
            job_type = "Default"

        # Initialize SageMaker session if needed (method from ModelBuilder mixin)
        if hasattr(self, '_init_sagemaker_session_if_does_not_exist'):
            self._init_sagemaker_session_if_does_not_exist()

        # Create inference recommendations job
        ret_name = self.sagemaker_session.create_inference_recommendations_job(
            role=getattr(self, 'role_arn', None),
            job_name=job_name,
            job_type=job_type,
            job_duration_in_seconds=job_duration_in_seconds,
            model_name=getattr(self, 'model_name', None),
            model_package_version_arn=getattr(self, "model_package_arn", None),
            framework=framework,
            framework_version=framework_version,
            sample_payload_url=sample_payload_url,
            supported_content_types=supported_content_types,
            supported_instance_types=supported_instance_types,
            endpoint_configurations=endpoint_configurations,
            traffic_pattern=traffic_pattern,
            stopping_conditions=stopping_conditions,
            resource_limit=resource_limit,
        )

        # Wait for job completion and store results
        self.inference_recommender_job_results = (
            self.sagemaker_session.wait_for_inference_recommendations_job(
                ret_name, log_level=log_level
            )
        )
        self.inference_recommendations = self.inference_recommender_job_results.get(
            "InferenceRecommendations", []
        )

        return self

    def _update_params(self, **kwargs) -> Optional[Tuple[str, int]]:
        """Update deployment parameters based on inference recommendations.
        
        Processes inference recommendation ID or right-size results to determine
        optimal instance type and count for model deployment.
        
        Args:
            **kwargs: Deployment parameters including instance_type, initial_instance_count,
                     inference_recommendation_id, etc.
                     
        Returns:
            Tuple of (instance_type, initial_instance_count) if recommendations found,
            otherwise None to use provided parameters.
        """
        instance_type = kwargs.get("instance_type")
        initial_instance_count = kwargs.get("initial_instance_count")
        accelerator_type = kwargs.get("accelerator_type")
        async_inference_config = kwargs.get("async_inference_config")
        serverless_inference_config = kwargs.get("serverless_inference_config")
        explainer_config = kwargs.get("explainer_config")
        inference_recommendation_id = kwargs.get("inference_recommendation_id")
        inference_recommender_job_results = kwargs.get("inference_recommender_job_results")
        
        inference_recommendation = None
        
        if inference_recommendation_id is not None:
            inference_recommendation = self._update_params_for_recommendation_id(
                instance_type=instance_type,
                initial_instance_count=initial_instance_count,
                accelerator_type=accelerator_type,
                async_inference_config=async_inference_config,
                serverless_inference_config=serverless_inference_config,
                inference_recommendation_id=inference_recommendation_id,
                explainer_config=explainer_config,
            )
        elif inference_recommender_job_results is not None:
            inference_recommendation = self._update_params_for_right_size(
                instance_type,
                initial_instance_count,
                accelerator_type,
                serverless_inference_config,
                async_inference_config,
                explainer_config,
            )

        return (
            inference_recommendation
            if inference_recommendation
            else (instance_type, initial_instance_count)
        )

    def _update_params_for_right_size(
        self,
        instance_type: Optional[str] = None,
        initial_instance_count: Optional[int] = None,
        accelerator_type: Optional[str] = None,
        serverless_inference_config: Optional[Any] = None,
        async_inference_config: Optional[Any] = None,
        explainer_config: Optional[Any] = None,
    ) -> Optional[Tuple[str, int]]:
        """Validates that Inference Recommendation parameters can be used in `model.deploy()`

        Args:
            instance_type (str): The initial number of instances to run
                in the ``Endpoint`` created from this ``Model``. If not using
                serverless inference or the model has not called ``right_size()``,
                then it need to be a number larger or equals
                to 1 (default: None)
            initial_instance_count (int):The EC2 instance type to deploy this Model to.
                For example, 'ml.p2.xlarge', or 'local' for local mode. If not using
                serverless inference or the model has not called ``right_size()``,
                then it is required to deploy a model.
                (default: None)
            accelerator_type (str): whether accelerator_type has been passed into `model.deploy()`.
            serverless_inference_config (sagemaker.serve.serverless.ServerlessInferenceConfig)):
                whether serverless_inference_config has been passed into `model.deploy()`.
            async_inference_config (sagemaker.model_monitor.AsyncInferenceConfig):
                whether async_inference_config has been passed into `model.deploy()`.
            explainer_config (sagemaker.explainer.ExplainerConfig): whether explainer_config
                has been passed into `model.deploy()`.

        Returns:
            (string, int) or None: Top instance_type and associated initial_instance_count
            if self.inference_recommender_job_results has been generated. Otherwise, return None.
        """
        if accelerator_type:
            raise ValueError("accelerator_type is not compatible with right_size().")
        if instance_type or initial_instance_count:
            logger.warning(
                "instance_type or initial_instance_count specified."
                "Overriding right_size() recommendations."
            )
            return None
        if async_inference_config:
            logger.warning(
                "async_inference_config is specified. Overriding right_size() recommendations."
            )
            return None
        if serverless_inference_config:
            logger.warning(
                "serverless_inference_config is specified. Overriding right_size() recommendations."
            )
            return None
        if explainer_config:
            logger.warning(
                "explainer_config is specified. Overriding right_size() recommendations."
            )
            return None

        return self._filter_recommendations_for_realtime()

    def _update_params_for_recommendation_id(
        self,
        instance_type: Optional[str],
        initial_instance_count: Optional[int],
        accelerator_type: Optional[str],
        async_inference_config: Optional[Any],
        serverless_inference_config: Optional[Any],
        inference_recommendation_id: str,
        explainer_config: Optional[Any],
    ) -> Tuple[str, int]:
        """Update parameters with inference recommendation results.

        Args:
            instance_type (str): The EC2 instance type to deploy this Model to.
                For example, 'ml.p2.xlarge', or 'local' for local mode. If not using
                serverless inference, then it is required to deploy a model.
            initial_instance_count (int): The initial number of instances to run
                in the ``Endpoint`` created from this ``Model``. If not using
                serverless inference, then it need to be a number larger or equals
                to 1.
            accelerator_type (str): Type of Elastic Inference accelerator to
                deploy this model for model loading and inference, for example,
                'ml.eia1.medium'. If not specified, no Elastic Inference
                accelerator will be attached to the endpoint. For more
                information:
                https://docs.aws.amazon.com/sagemaker/latest/dg/ei.html
            async_inference_config (sagemaker.model_monitor.AsyncInferenceConfig): Specifies
                configuration related to async endpoint. Use this configuration when trying
                to create async endpoint and make async inference. If empty config object
                passed through, will use default config to deploy async endpoint. Deploy a
                real-time endpoint if it's None.
            serverless_inference_config (sagemaker.serve.serverless.ServerlessInferenceConfig):
                Specifies configuration related to serverless endpoint. Use this configuration
                when trying to create serverless endpoint and make serverless inference. If
                empty object passed through, will use pre-defined values in
                ``ServerlessInferenceConfig`` class to deploy serverless endpoint. Deploy an
                instance based endpoint if it's None.
            inference_recommendation_id (str): The recommendation id which specifies
                the recommendation you picked from inference recommendation job
                results and would like to deploy the model and endpoint with
                recommended parameters.
            explainer_config (sagemaker.explainer.ExplainerConfig): Specifies online explainability
                configuration for use with Amazon SageMaker Clarify. Default: None.
        Raises:
            ValueError: If arguments combination check failed in these circumstances:
                - If only one of instance type or instance count specified or
                - If recommendation id does not follow the required format or
                - If recommendation id is not valid or
                - If inference recommendation id is specified along with incompatible parameters
        Returns:
            (string, int): instance type and associated instance count from selected
            inference recommendation id if arguments combination check passed.
        """

        if instance_type is not None and initial_instance_count is not None:
            logger.warning(
                "Both instance_type and initial_instance_count are specified,"
                "overriding the recommendation result."
            )
            return (instance_type, initial_instance_count)

        # Validate non-compatible parameters with recommendation id
        if accelerator_type is not None:
            raise ValueError("accelerator_type is not compatible with inference_recommendation_id.")
        if async_inference_config is not None:
            raise ValueError(
                "async_inference_config is not compatible with inference_recommendation_id."
            )
        if serverless_inference_config is not None:
            raise ValueError(
                "serverless_inference_config is not compatible with inference_recommendation_id."
            )
        if explainer_config is not None:
            raise ValueError("explainer_config is not compatible with inference_recommendation_id.")

        # Validate recommendation ID format
        if not re.match(r"[a-zA-Z0-9](-*[a-zA-Z0-9]){0,63}\/\w{8}$", inference_recommendation_id):
            raise ValueError(
                f"Invalid inference_recommendation_id format: {inference_recommendation_id}. "
                f"Expected format: <job-or-model-name>/<8-character-id>"
            )
        job_or_model_name = inference_recommendation_id.split("/")[0]

        sage_client = self.sagemaker_session.sagemaker_client
        # Get recommendation from right size job and model
        (
            right_size_recommendation,
            model_recommendation,
            right_size_job_res,
        ) = self._get_recommendation(
            sage_client=sage_client,
            job_or_model_name=job_or_model_name,
            inference_recommendation_id=inference_recommendation_id,
        )

        # Update params based on model recommendation
        if model_recommendation:
            if initial_instance_count is None:
                raise ValueError(
                    "Must specify initial_instance_count when using model recommendation ID."
                )
            # Update environment variables if they exist
            env_vars = getattr(self, 'env_vars', {})
            env_vars.update(model_recommendation.get("Environment", {}))
            instance_type = model_recommendation["InstanceType"]
            return (instance_type, initial_instance_count)

        # Update params based on default inference recommendation
        if bool(instance_type) != bool(initial_instance_count):
            raise ValueError(
                "instance_type and initial_instance_count must both be specified together "
                "to override recommendation, or both omitted to use recommendation values."
            )
            
        input_config = right_size_job_res["InputConfig"]
        model_config = right_size_recommendation["ModelConfiguration"]
        envs = model_config.get("EnvironmentParameters")
        
        # Update environment variables from recommendation
        recommend_envs = {}
        if envs:
            for env in envs:
                recommend_envs[env["Key"]] = env["Value"]
                
        # Safely update env_vars
        current_env_vars = getattr(self, 'env_vars', {})
        current_env_vars.update(recommend_envs)

        # Update params with non-compilation recommendation results
        if (
            "InferenceSpecificationName" not in model_config
            and "CompilationJobName" not in model_config
        ):

            if "ModelPackageVersionArn" in input_config:
                modelpkg_res = sage_client.describe_model_package(
                    ModelPackageName=input_config["ModelPackageVersionArn"]
                )
                self.s3_model_data_url = modelpkg_res["InferenceSpecification"]["Containers"][0][
                    "ModelDataUrl"
                ]
                self.image_uri = modelpkg_res["InferenceSpecification"]["Containers"][0]["Image"]
            elif "ModelName" in input_config:
                model_res = sage_client.describe_model(ModelName=input_config["ModelName"])
                self.s3_model_data_url = model_res["PrimaryContainer"]["ModelDataUrl"]
                self.image_uri = model_res["PrimaryContainer"]["Image"]
        else:
            if "InferenceSpecificationName" in model_config:
                modelpkg_res = sage_client.describe_model_package(
                    ModelPackageName=input_config["ModelPackageVersionArn"]
                )
                self.s3_model_data_url = modelpkg_res["AdditionalInferenceSpecificationDefinition"][
                    "Containers"
                ][0]["ModelDataUrl"]
                self.image_uri = modelpkg_res["AdditionalInferenceSpecificationDefinition"][
                    "Containers"
                ][0]["Image"]
            elif "CompilationJobName" in model_config:
                compilation_res = sage_client.describe_compilation_job(
                    CompilationJobName=model_config["CompilationJobName"]
                )
                self.s3_model_data_url = compilation_res["ModelArtifacts"]["S3ModelArtifacts"]
                self.image_uri = compilation_res["InferenceImage"]

        instance_type = right_size_recommendation["EndpointConfiguration"]["InstanceType"]
        initial_instance_count = right_size_recommendation["EndpointConfiguration"][
            "InitialInstanceCount"
        ]

        return (instance_type, initial_instance_count)

    def _convert_to_endpoint_configurations_json(
        self, hyperparameter_ranges: Optional[List[Dict[str, CategoricalParameter]]]
    ) -> Optional[List[Dict[str, Any]]]:
        """Convert hyperparameter ranges to endpoint configurations for Advanced jobs.
        
        Args:
            hyperparameter_ranges: List of hyperparameter range dictionaries
            
        Returns:
            List of endpoint configuration dictionaries, or None if no ranges provided
            
        Raises:
            ValueError: If instance_types not specified in hyperparameter ranges
        """
        if not hyperparameter_ranges:
            return None

        endpoint_configurations_to_json = []
        for parameter_range in hyperparameter_ranges:
            if not parameter_range.get("instance_types"):
                raise ValueError(
                    "instance_types must be defined as a hyperparameter range for Advanced jobs"
                )
            parameter_range = parameter_range.copy()
            instance_types = parameter_range.get("instance_types").values
            parameter_range.pop("instance_types")

            for instance_type in instance_types:
                parameter_ranges = [
                    {"Name": name, "Value": param.values} for name, param in parameter_range.items()
                ]
                endpoint_configurations_to_json.append(
                    {
                        "EnvironmentParameterRanges": {
                            "CategoricalParameterRanges": parameter_ranges
                        },
                        "InstanceType": instance_type,
                    }
                )

        return endpoint_configurations_to_json

    def _convert_to_traffic_pattern_json(
        self, traffic_type: Optional[str], phases: Optional[List[Phase]]
    ) -> Optional[Dict[str, Any]]:
        """Convert traffic pattern parameters for Advanced jobs.
        
        Args:
            traffic_type: Type of traffic pattern (defaults to "PHASES")
            phases: List of Phase objects defining load test pattern
            
        Returns:
            Traffic pattern dictionary, or None if no phases provided
        """
        if not phases:
            return None
        return {
            "Phases": [phase.to_json for phase in phases],
            "TrafficType": traffic_type if traffic_type else "PHASES",
        }

    def _convert_to_resource_limit_json(
        self, max_tests: Optional[int], max_parallel_tests: Optional[int]
    ) -> Optional[Dict[str, int]]:
        """Convert resource limit parameters for Advanced jobs.
        
        Args:
            max_tests: Maximum number of tests to run
            max_parallel_tests: Maximum number of parallel tests
            
        Returns:
            Resource limit dictionary, or None if no limits specified
        """
        if not max_tests and not max_parallel_tests:
            return None
        resource_limit = {}
        if max_tests:
            resource_limit["MaxNumberOfTests"] = max_tests
        if max_parallel_tests:
            resource_limit["MaxParallelOfTests"] = max_parallel_tests
        return resource_limit

    def _convert_to_stopping_conditions_json(
        self, 
        max_invocations: Optional[int], 
        model_latency_thresholds: Optional[List[ModelLatencyThreshold]]
    ) -> Optional[Dict[str, Any]]:
        """Convert stopping condition parameters for Advanced jobs.
        
        Args:
            max_invocations: Maximum number of invocations per minute
            model_latency_thresholds: List of latency threshold requirements
            
        Returns:
            Stopping conditions dictionary, or None if no conditions specified
        """
        if not max_invocations and not model_latency_thresholds:
            return None
        stopping_conditions = {}
        if max_invocations:
            stopping_conditions["MaxInvocations"] = max_invocations
        if model_latency_thresholds:
            stopping_conditions["ModelLatencyThresholds"] = [
                threshold.to_json for threshold in model_latency_thresholds
            ]
        return stopping_conditions

    def _get_recommendation(
        self, sage_client: Any, job_or_model_name: str, inference_recommendation_id: str
    ) -> Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]], Optional[Dict[str, Any]]]:
        """Retrieve recommendation from right-size job or model.
        
        Args:
            sage_client: SageMaker client for API calls
            job_or_model_name: Name of the job or model
            inference_recommendation_id: ID of the specific recommendation
            
        Returns:
            Tuple of (right_size_recommendation, model_recommendation, right_size_job_res)
            
        Raises:
            ValueError: If recommendation ID is not found in any source
        """
        right_size_recommendation, model_recommendation, right_size_job_res = None, None, None
        
        # Try to get recommendation from right-size job first
        right_size_recommendation, right_size_job_res = self._get_right_size_recommendation(
            sage_client=sage_client,
            job_or_model_name=job_or_model_name,
            inference_recommendation_id=inference_recommendation_id,
        )
        
        # If not found in job, try model recommendations
        if right_size_recommendation is None:
            model_recommendation = self._get_model_recommendation(
                sage_client=sage_client,
                job_or_model_name=job_or_model_name,
                inference_recommendation_id=inference_recommendation_id,
            )
            if model_recommendation is None:
                raise ValueError(
                    f"Recommendation ID '{inference_recommendation_id}' not found in "
                    f"job '{job_or_model_name}' or associated model recommendations"
                )

        return right_size_recommendation, model_recommendation, right_size_job_res

    def _get_right_size_recommendation(
        self,
        sage_client: Any,
        job_or_model_name: str,
        inference_recommendation_id: str,
    ) -> Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]]]:
        """Get recommendation from right-size job.
        
        Args:
            sage_client: SageMaker client
            job_or_model_name: Name of the inference recommendations job
            inference_recommendation_id: Specific recommendation ID to find
            
        Returns:
            Tuple of (recommendation, job_results) or (None, None) if not found
        """
        right_size_recommendation, right_size_job_res = None, None
        try:
            right_size_job_res = sage_client.describe_inference_recommendations_job(
                JobName=job_or_model_name
            )
            if right_size_job_res:
                right_size_recommendation = self._search_recommendation(
                    recommendation_list=right_size_job_res.get("InferenceRecommendations", []),
                    inference_recommendation_id=inference_recommendation_id,
                )
        except sage_client.exceptions.ResourceNotFound:
            pass

        return right_size_recommendation, right_size_job_res

    def _get_model_recommendation(
        self,
        sage_client: Any,
        job_or_model_name: str,
        inference_recommendation_id: str,
    ) -> Optional[Dict[str, Any]]:
        """Get recommendation from model deployment recommendations.
        
        Args:
            sage_client: SageMaker client
            job_or_model_name: Name of the model
            inference_recommendation_id: Specific recommendation ID to find
            
        Returns:
            Model recommendation dictionary or None if not found
        """
        model_recommendation = None
        try:
            model_res = sage_client.describe_model(ModelName=job_or_model_name)
            if model_res:
                deployment_rec = model_res.get("DeploymentRecommendation", {})
                realtime_recs = deployment_rec.get("RealTimeInferenceRecommendations", [])
                model_recommendation = self._search_recommendation(
                    recommendation_list=realtime_recs,
                    inference_recommendation_id=inference_recommendation_id,
                )
        except sage_client.exceptions.ResourceNotFound:
            pass

        return model_recommendation

    def _search_recommendation(
        self, recommendation_list: List[Dict[str, Any]], inference_recommendation_id: str
    ) -> Optional[Dict[str, Any]]:
        """Search for specific recommendation by ID.
        
        Args:
            recommendation_list: List of recommendation dictionaries
            inference_recommendation_id: ID to search for
            
        Returns:
            Matching recommendation dictionary or None if not found
        """
        return next(
            (
                rec
                for rec in recommendation_list
                if rec.get("RecommendationId") == inference_recommendation_id
            ),
            None,
        )

    def _filter_recommendations_for_realtime(self) -> Tuple[Optional[str], Optional[int]]:
        """Filter recommendations to find real-time (non-serverless) instance.
        
        Returns:
            Tuple of (instance_type, initial_instance_count) for first real-time
            recommendation found, or (None, None) if none found.
            
        Note:
            TODO: Integrate right_size + deploy with serverless support
        """
        instance_type = None
        initial_instance_count = None
        
        inference_recommendations = getattr(self, 'inference_recommendations', [])
        for recommendation in inference_recommendations:
            endpoint_config = recommendation.get("EndpointConfiguration", {})
            if "ServerlessConfig" not in endpoint_config:
                instance_type = endpoint_config.get("InstanceType")
                initial_instance_count = endpoint_config.get("InitialInstanceCount")
                break
                
        return (instance_type, initial_instance_count)



"""Placeholder docstring"""

from __future__ import absolute_import

CONFIG_TEMPLATE = """name: \"model\"
backend: \"python\"

input [
  {{
    name: \"{input_name}\"
    data_type: {input_dtype}
    dims: {input_shape}
  }}
]

output [
  {{
    name: \"{output_name}\"
    data_type: {output_dtype}
    dims: {output_shape}
  }}
]

instance_group [
    {{
    kind: {hardware_type}
    }}
]

parameters: {{
  key: \"EXECUTION_ENV_PATH\",
  value: {{string_value: \"$$TRITON_MODEL_DIRECTORY/triton_env.tar.gz\"}}
}}
"""



"""Module for artifacts preparation for tensorflow_serving"""

from __future__ import absolute_import
from pathlib import Path
import shutil
from typing import List, Dict, Any

from sagemaker.serve.model_format.mlflow.utils import (
    _get_saved_model_path_for_tensorflow_and_keras_flavor,
    _move_contents,
)
from sagemaker.serve.detector.dependency_manager import capture_dependencies
from sagemaker.serve.validations.check_integrity import (
    generate_secret_key,
    compute_hash,
)
from sagemaker.core.remote_function.core.serialization import _MetaData


def prepare_for_tf_serving(
    model_path: str,
    shared_libs: List[str],
    dependencies: Dict[str, Any],
) -> str:
    """Prepares the model for serving.

    Args:
        model_path (str): Path to the model directory.
        shared_libs (List[str]): List of shared libraries.
        dependencies (Dict[str, Any]): Dictionary of dependencies.

    Returns:
        str: Secret key.
    """

    _model_path = Path(model_path)
    if not _model_path.exists():
        _model_path.mkdir()
    elif not _model_path.is_dir():
        raise Exception("model_dir is not a valid directory")

    code_dir = _model_path.joinpath("code")
    code_dir.mkdir(exist_ok=True)
    shutil.copy2(Path(__file__).parent.joinpath("inference.py"), code_dir)

    shared_libs_dir = _model_path.joinpath("shared_libs")
    shared_libs_dir.mkdir(exist_ok=True)
    for shared_lib in shared_libs:
        shutil.copy2(Path(shared_lib), shared_libs_dir)

    capture_dependencies(dependencies=dependencies, work_dir=code_dir)

    saved_model_bundle_dir = _model_path.joinpath("1")
    saved_model_bundle_dir.mkdir(exist_ok=True)
    mlflow_saved_model_dir = _get_saved_model_path_for_tensorflow_and_keras_flavor(model_path)
    if not mlflow_saved_model_dir:
        raise ValueError("SavedModel is not found for Tensorflow or Keras flavor.")
    _move_contents(src_dir=mlflow_saved_model_dir, dest_dir=saved_model_bundle_dir)
    
    secret_key = generate_secret_key()
    with open(str(code_dir.joinpath("serve.pkl")), "rb") as f:
        buffer = f.read()
    hash_value = compute_hash(buffer=buffer, secret_key=secret_key)
    with open(str(code_dir.joinpath("metadata.json")), "wb") as metadata:
        metadata.write(_MetaData(hash_value).to_json())

    return secret_key



"""Summary of MyModule.

Extended discussion of my module.
"""

from __future__ import absolute_import
import os
from pathlib import Path
import shutil
from typing import List

from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.detector.dependency_manager import capture_dependencies
from sagemaker.serve.validations.check_integrity import (
    generate_secret_key,
    compute_hash,
)
from sagemaker.core.remote_function.core.serialization import _MetaData
from sagemaker.serve.spec.inference_base import CustomOrchestrator, AsyncCustomOrchestrator


def prepare_for_smd(
    model_path: str,
    shared_libs: List[str],
    dependencies: dict,
    inference_spec: InferenceSpec = None,
) -> str:
    """Prepares artifacts for SageMaker model deployment.

    Args:to
        model_path (str) : Argument
        shared_libs (List[]) : Argument
        dependencies (dict) : Argument
        inference_spec (InferenceSpec, optional) : Argument
            (default is None)

    Returns:
        ( str ) :

    """
    model_path = Path(model_path)
    if not model_path.exists():
        model_path.mkdir()
    elif not model_path.is_dir():
        raise Exception("model_dir is not a valid directory")

    if inference_spec and isinstance(inference_spec, InferenceSpec):
        inference_spec.prepare(str(model_path))

    code_dir = model_path.joinpath("code")
    code_dir.mkdir(exist_ok=True)

    if inference_spec and isinstance(inference_spec, (CustomOrchestrator, AsyncCustomOrchestrator)):
        shutil.copy2(Path(__file__).parent.joinpath("custom_execution_inference.py"), code_dir)
        os.rename(
            str(code_dir.joinpath("custom_execution_inference.py")),
            str(code_dir.joinpath("inference.py")),
        )

    shared_libs_dir = model_path.joinpath("shared_libs")
    shared_libs_dir.mkdir(exist_ok=True)
    for shared_lib in shared_libs:
        shutil.copy2(Path(shared_lib), shared_libs_dir)

    capture_dependencies(dependencies=dependencies, work_dir=code_dir)

    secret_key = generate_secret_key()
    with open(str(code_dir.joinpath("serve.pkl")), "rb") as f:
        buffer = f.read()
    hash_value = compute_hash(buffer=buffer, secret_key=secret_key)
    with open(str(code_dir.joinpath("metadata.json")), "wb") as metadata:
        metadata.write(_MetaData(hash_value).to_json())

    return secret_key



"""This module is for SageMaker inference.py."""

from __future__ import absolute_import
import os
import io
import cloudpickle
import shutil
import platform
from pathlib import Path
from functools import partial
from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.validations.check_integrity import perform_integrity_check
import logging

logger = logging.getLogger(__name__)

inference_spec = None
schema_builder = None
SHARED_LIBS_DIR = Path(__file__).parent.parent.joinpath("shared_libs")
SERVE_PATH = Path(__file__).parent.joinpath("serve.pkl")
METADATA_PATH = Path(__file__).parent.joinpath("metadata.json")


def model_fn(model_dir, context=None):
    """Overrides default method for loading a model"""
    shared_libs_path = Path(model_dir + "/shared_libs")

    if shared_libs_path.exists():
        # before importing, place dynamic linked libraries in shared lib path
        shutil.copytree(shared_libs_path, "/lib", dirs_exist_ok=True)

    serve_path = Path(__file__).parent.joinpath("serve.pkl")
    with open(str(serve_path), mode="rb") as file:
        global inference_spec, schema_builder
        obj = cloudpickle.load(file)
        if isinstance(obj[0], InferenceSpec):
            inference_spec, schema_builder = obj

    if inference_spec:
        return partial(inference_spec.invoke, model=inference_spec.load(model_dir))


def input_fn(input_data, content_type, context=None):
    """Deserializes the bytes that were received from the model server"""
    try:
        if hasattr(schema_builder, "custom_input_translator"):
            deserialized_data = schema_builder.custom_input_translator.deserialize(
                (
                    io.BytesIO(input_data.encode("utf-8"))
                    if not any(
                        [
                            isinstance(input_data, bytes),
                            isinstance(input_data, bytearray),
                        ]
                    )
                    else io.BytesIO(input_data)
                ),
                content_type,
            )
        else:
            deserialized_data = schema_builder.input_deserializer.deserialize(
                (
                    io.BytesIO(input_data.encode("utf-8"))
                    if not any(
                        [
                            isinstance(input_data, bytes),
                            isinstance(input_data, bytearray),
                        ]
                    )
                    else io.BytesIO(input_data)
                ),
                content_type[0],
            )

        # Check if preprocess method is defined and call it
        if hasattr(inference_spec, "preprocess"):
            preprocessed = inference_spec.preprocess(deserialized_data)
            if preprocessed is not None:
                return preprocessed

        return deserialized_data
    except Exception as e:
        logger.error("Encountered error: %s in deserialize_response." % e)
        raise Exception("Encountered error in deserialize_request.") from e


def predict_fn(input_data, predict_callable, context=None):
    """Invokes the model that is taken in by model server"""
    return predict_callable(input_data)


def output_fn(predictions, accept_type, context=None):
    """Prediction is serialized to bytes and sent back to the customer"""
    try:
        if hasattr(inference_spec, "postprocess"):
            postprocessed = inference_spec.postprocess(predictions)
            if postprocessed is not None:
                predictions = postprocessed
        if hasattr(schema_builder, "custom_output_translator"):
            return schema_builder.custom_output_translator.serialize(predictions, accept_type)
        else:
            return schema_builder.output_serializer.serialize(predictions)
    except Exception as e:
        logger.error("Encountered error: %s in serialize_response." % e)
        raise Exception("Encountered error in serialize_response.") from e


def _run_preflight_diagnostics():
    _py_vs_parity_check()
    _pickle_file_integrity_check()


def _py_vs_parity_check():
    container_py_vs = platform.python_version()
    local_py_vs = os.getenv("LOCAL_PYTHON")

    if not local_py_vs or container_py_vs.split(".")[1] != local_py_vs.split(".")[1]:
        logger.warning(
            f"The local python version {local_py_vs} differs from the python version "
            f"{container_py_vs} on the container. Please align the two to avoid unexpected behavior"
        )


def _pickle_file_integrity_check():
    with open(SERVE_PATH, "rb") as f:
        buffer = f.read()

    perform_integrity_check(buffer=buffer, metadata_path=METADATA_PATH)


# on import, execute
_run_preflight_diagnostics()



"""Placeholder docstring"""

from __future__ import absolute_import
from pathlib import Path

import logging
from typing import Type

from sagemaker.serve.model_server.tei.server import SageMakerTeiServing
from sagemaker.serve.model_server.tensorflow_serving.server import SageMakerTensorflowServing
from sagemaker.core.helper.session_helper import Session
from sagemaker.serve.utils.types import ModelServer
from sagemaker.serve.spec.inference_spec import InferenceSpec
from sagemaker.serve.model_server.triton.server import SageMakerTritonServer
from sagemaker.serve.model_server.torchserve.server import SageMakerTorchServe
from sagemaker.serve.model_server.djl_serving.server import SageMakerDjlServing
from sagemaker.serve.model_server.tgi.server import SageMakerTgiServing
from sagemaker.serve.model_server.multi_model_server.server import SageMakerMultiModelServer
from sagemaker.serve.model_server.smd.server import SageMakerSmdServer

logger = logging.getLogger(__name__)


class SageMakerEndpointMode(
    SageMakerTorchServe,
    SageMakerTritonServer,
    SageMakerDjlServing,
    SageMakerTgiServing,
    SageMakerMultiModelServer,
    SageMakerTensorflowServing,
    SageMakerSmdServer,
):
    """Holds the required method to deploy a model to a SageMaker Endpoint"""

    def __init__(self, inference_spec: Type[InferenceSpec], model_server: ModelServer):
        super().__init__()
        # pylint: disable=bad-super-call
        super(SageMakerTritonServer, self).__init__()

        self.inference_spec = inference_spec
        self.model_server = model_server

        self._tei_serving = SageMakerTeiServing()

    def load(self, model_path: str):
        """Placeholder docstring"""
        path = Path(model_path)
        if not path.exists():
            raise Exception("model_path does not exist")
        if not path.is_dir():
            raise Exception("model_path is not a valid directory")

        model_dir = path.joinpath("model")
        return self.inference_spec.model_fn(str(model_dir))

    def prepare(
        self,
        model_path: str,
        secret_key: str,
        s3_model_data_url: str = None,
        sagemaker_session: Session = None,
        image: str = None,
        jumpstart: bool = False,
        should_upload_artifacts: bool = False,
    ):
        """Placeholder docstring"""
        try:
            sagemaker_session = sagemaker_session or Session()
        except Exception as e:
            raise Exception(
                "Failed to setup default SageMaker session. Please allow a default "
                + "session to be created or supply `sagemaker_session` into @serve.invoke."
            ) from e

        upload_artifacts = None, None
        if self.model_server == ModelServer.TORCHSERVE:
            upload_artifacts = self._upload_torchserve_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                secret_key=secret_key,
                s3_model_data_url=s3_model_data_url,
                image=image,
                should_upload_artifacts=True,
            )

        if self.model_server == ModelServer.TRITON:
            upload_artifacts = self._upload_triton_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                secret_key=secret_key,
                s3_model_data_url=s3_model_data_url,
                image=image,
                should_upload_artifacts=True,
            )

        if self.model_server == ModelServer.DJL_SERVING:
            upload_artifacts = self._upload_djl_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                s3_model_data_url=s3_model_data_url,
                image=image,
                should_upload_artifacts=True,
            )

        if self.model_server == ModelServer.TENSORFLOW_SERVING:
            upload_artifacts = self._upload_tensorflow_serving_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                secret_key=secret_key,
                s3_model_data_url=s3_model_data_url,
                image=image,
                should_upload_artifacts=True,
            )

        # By default, we do not want to upload artifacts in S3 for the below server.
        # In Case of Optimization, artifacts need to be uploaded into s3.
        # In that case, `should_upload_artifacts` arg needs to come from
        # the caller of prepare.

        if self.model_server == ModelServer.TGI:
            upload_artifacts = self._upload_tgi_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                s3_model_data_url=s3_model_data_url,
                image=image,
                jumpstart=jumpstart,
                should_upload_artifacts=should_upload_artifacts,
            )

        if self.model_server == ModelServer.MMS:
            upload_artifacts = self._upload_server_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                s3_model_data_url=s3_model_data_url,
                secret_key=secret_key,
                image=image,
                should_upload_artifacts=should_upload_artifacts,
            )

        if self.model_server == ModelServer.TEI:
            upload_artifacts = self._tei_serving._upload_tei_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                s3_model_data_url=s3_model_data_url,
                image=image,
                should_upload_artifacts=should_upload_artifacts,
            )

        if self.model_server == ModelServer.SMD:
            upload_artifacts = self._upload_smd_artifacts(
                model_path=model_path,
                sagemaker_session=sagemaker_session,
                secret_key=secret_key,
                s3_model_data_url=s3_model_data_url,
                image=image,
                should_upload_artifacts=True,
            )

        if upload_artifacts or isinstance(self.model_server, ModelServer):
            return upload_artifacts

        raise ValueError("%s model server is not supported" % self.model_server)
